# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module tests fitting and model evaluation with various inputs
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.modeling import fitting, models
from astropy.modeling.core import Fittable1DModel, FittableModel, Model
from astropy.modeling.parameters import Parameter
from astropy.utils.compat.optional_deps import HAS_SCIPY

model1d_params = [
    (models.Polynomial1D, [2]),
    (models.Legendre1D, [2]),
    (models.Chebyshev1D, [2]),
    (models.Shift, [2]),
    (models.Scale, [2]),
]

model2d_params = [
    (models.Polynomial2D, [2]),
    (models.Legendre2D, [1, 2]),
    (models.Chebyshev2D, [1, 2]),
]

fitters = [
    fitting.LevMarLSQFitter,
    fitting.TRFLSQFitter,
    fitting.LMLSQFitter,
    fitting.DogBoxLSQFitter,
]

fitters_bounds = [
    fitting.LevMarLSQFitter,
    fitting.TRFLSQFitter,
    fitting.DogBoxLSQFitter,
]


class TestInputType:
    """
    This class tests that models accept numbers, lists and arrays.

    Add new models to one of the lists above to test for this.
    """

    def setup_class(self):
        self.x = 5.3
        self.y = 6.7
        self.x1 = np.arange(1, 10, 0.1)
        self.y1 = np.arange(1, 10, 0.1)
        self.y2, self.x2 = np.mgrid[:10, :8]

    @pytest.mark.parametrize(("model", "params"), model1d_params)
    def test_input1D(self, model, params):
        m = model(*params)
        m(self.x)
        m(self.x1)
        m(self.x2)

    @pytest.mark.parametrize(("model", "params"), model2d_params)
    def test_input2D(self, model, params):
        m = model(*params)
        m(self.x, self.y)
        m(self.x1, self.y1)
        m(self.x2, self.y2)


class TestFitting:
    """Test various input options to fitting routines."""

    def setup_class(self):
        self.x1 = np.arange(10)
        self.y, self.x = np.mgrid[:10, :10]

    def test_linear_fitter_1set(self):
        """1 set 1D x, 1pset"""

        expected = np.array([0, 1, 1, 1])
        p1 = models.Polynomial1D(3)
        p1.parameters = [0, 1, 1, 1]
        y1 = p1(self.x1)
        pfit = fitting.LinearLSQFitter()
        model = pfit(p1, self.x1, y1)
        assert_allclose(model.parameters, expected, atol=10 ** (-7))

    def test_linear_fitter_Nset(self):
        """1 set 1D x, 2 sets 1D y, 2 param_sets"""

        expected = np.array([[0, 0], [1, 1], [2, 2], [3, 3]])
        p1 = models.Polynomial1D(3, n_models=2)
        p1.parameters = [0.0, 0.0, 1.0, 1.0, 2.0, 2.0, 3.0, 3.0]
        params = {}
        for i in range(4):
            params[p1.param_names[i]] = [i, i]
        p1 = models.Polynomial1D(3, model_set_axis=0, **params)
        y1 = p1(self.x1, model_set_axis=False)
        pfit = fitting.LinearLSQFitter()
        model = pfit(p1, self.x1, y1)
        assert_allclose(model.param_sets, expected, atol=10 ** (-7))

    def test_linear_fitter_1dcheb(self):
        """1 pset, 1 set 1D x, 1 set 1D y, Chebyshev 1D polynomial"""

        expected = np.array(
            [
                [
                    2817.2499999999995,
                    4226.6249999999991,
                    1680.7500000000009,
                    273.37499999999926,
                ]
            ]
        ).T
        ch1 = models.Chebyshev1D(3)
        ch1.parameters = [0, 1, 2, 3]
        y1 = ch1(self.x1)
        pfit = fitting.LinearLSQFitter()
        model = pfit(ch1, self.x1, y1)
        assert_allclose(model.param_sets, expected, atol=10 ** (-2))

    def test_linear_fitter_1dlegend(self):
        """
        1 pset, 1 set 1D x, 1 set 1D y, Legendre 1D polynomial
        """

        expected = np.array(
            [
                [
                    1925.5000000000011,
                    3444.7500000000005,
                    1883.2500000000014,
                    364.4999999999996,
                ]
            ]
        ).T
        leg1 = models.Legendre1D(3)
        leg1.parameters = [1, 2, 3, 4]
        y1 = leg1(self.x1)
        pfit = fitting.LinearLSQFitter()
        model = pfit(leg1, self.x1, y1)
        assert_allclose(model.param_sets, expected, atol=10 ** (-12))

    def test_linear_fitter_1set2d(self):
        p2 = models.Polynomial2D(2)
        p2.parameters = [0, 1, 2, 3, 4, 5]
        expected = [0, 1, 2, 3, 4, 5]
        z = p2(self.x, self.y)
        pfit = fitting.LinearLSQFitter()
        model = pfit(p2, self.x, self.y, z)
        assert_allclose(model.parameters, expected, atol=10 ** (-12))
        assert_allclose(model(self.x, self.y), z, atol=10 ** (-12))

    def test_wrong_numpset(self):
        """
        A ValueError is raised if a 1 data set (1d x, 1d y) is fit
        with a model with multiple parameter sets.
        """

        MESSAGE = (
            r"Number of data sets .* is expected to equal the number of parameter sets"
        )
        with pytest.raises(ValueError, match=MESSAGE):
            p1 = models.Polynomial1D(5)
            y1 = p1(self.x1)
            p1 = models.Polynomial1D(5, n_models=2)
            pfit = fitting.LinearLSQFitter()
            pfit(p1, self.x1, y1)

    def test_wrong_pset(self):
        """A case of 1 set of x and multiple sets of y and parameters."""

        expected = np.array(
            [
                [1, 0],
                [1, 1],
                [1, 2],
                [1, 3],
                [1, 4],
                [1, 5],
            ]
        )
        p1 = models.Polynomial1D(5, n_models=2)
        params = {}
        for i in range(6):
            params[p1.param_names[i]] = [1, i]
        p1 = models.Polynomial1D(5, model_set_axis=0, **params)
        y1 = p1(self.x1, model_set_axis=False)
        pfit = fitting.LinearLSQFitter()
        model = pfit(p1, self.x1, y1)
        assert_allclose(model.param_sets, expected, atol=10 ** (-7))

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    @pytest.mark.parametrize("fitter", fitters_bounds)
    def test_nonlinear_lsqt_1set_1d(self, fitter):
        """1 set 1D x, 1 set 1D y, 1 pset NonLinearFitter"""
        fitter = fitter()

        g1 = models.Gaussian1D(10, mean=3, stddev=0.2)
        y1 = g1(self.x1)
        model = fitter(g1, self.x1, y1)
        assert_allclose(model.parameters, [10, 3, 0.2])

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    @pytest.mark.parametrize("fitter", fitters_bounds)
    def test_nonlinear_lsqt_Nset_1d(self, fitter):
        """1 set 1D x, 1 set 1D y, 2 param_sets, NonLinearFitter"""
        fitter = fitter()

        MESSAGE = r"Non-linear fitters can only fit one data set at a time"
        with pytest.raises(ValueError, match=MESSAGE):
            g1 = models.Gaussian1D(
                [10.2, 10], mean=[3, 3.2], stddev=[0.23, 0.2], n_models=2
            )
            y1 = g1(self.x1, model_set_axis=False)
            _ = fitter(g1, self.x1, y1)

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    @pytest.mark.parametrize("fitter", fitters_bounds)
    def test_nonlinear_lsqt_1set_2d(self, fitter):
        """1 set 2d x, 1set 2D y, 1 pset, NonLinearFitter"""
        fitter = fitter()

        g2 = models.Gaussian2D(
            10, x_mean=3, y_mean=4, x_stddev=0.3, y_stddev=0.2, theta=0
        )
        z = g2(self.x, self.y)
        model = fitter(g2, self.x, self.y, z)
        assert_allclose(model.parameters, [10, 3, 4, 0.3, 0.2, 0])

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    @pytest.mark.parametrize("fitter", fitters_bounds)
    def test_nonlinear_lsqt_Nset_2d(self, fitter):
        """1 set 2d x, 1set 2D y, 2 param_sets, NonLinearFitter"""
        fitter = fitter()

        MESSAGE = (
            r"Input argument .* does not have the correct dimensions in .* for a model"
            r" set with .*"
        )
        with pytest.raises(ValueError, match=MESSAGE):
            g2 = models.Gaussian2D(
                [10, 10],
                [3, 3],
                [4, 4],
                x_stddev=[0.3, 0.3],
                y_stddev=[0.2, 0.2],
                theta=[0, 0],
                n_models=2,
            )
            z = g2(self.x.ravel(), self.y.ravel())
            _ = fitter(g2, self.x, self.y, z)


class TestEvaluation:
    """
    Test various input options to model evaluation

    TestFitting actually covers evaluation of polynomials
    """

    def setup_class(self):
        self.x1 = np.arange(20)
        self.y, self.x = np.mgrid[:10, :10]

    def test_non_linear_NYset(self):
        """
        This case covers:
            N param sets , 1 set 1D x --> N 1D y data
        """

        g1 = models.Gaussian1D([10, 10], [3, 3], [0.2, 0.2], n_models=2)
        y1 = g1(self.x1, model_set_axis=False)
        assert np.all((y1[0, :] - y1[1, :]).nonzero() == np.array([]))

    def test_non_linear_NXYset(self):
        """
        This case covers: N param sets , N sets 1D x --> N N sets 1D y data
        """

        g1 = models.Gaussian1D([10, 10], [3, 3], [0.2, 0.2], n_models=2)
        xx = np.array([self.x1, self.x1])
        y1 = g1(xx)
        assert_allclose(y1[:, 0], y1[:, 1], atol=10 ** (-12))

    def test_p1_1set_1pset(self):
        """1 data set, 1 pset, Polynomial1D"""

        p1 = models.Polynomial1D(4)
        y1 = p1(self.x1)
        assert y1.shape == (20,)

    def test_p1_nset_npset(self):
        """N data sets, N param_sets, Polynomial1D"""

        p1 = models.Polynomial1D(4, n_models=2)
        y1 = p1(np.array([self.x1, self.x1]).T, model_set_axis=-1)
        assert y1.shape == (20, 2)
        assert_allclose(y1[0, :], y1[1, :], atol=10 ** (-12))

    def test_p2_1set_1pset(self):
        """1 pset, 1 2D data set, Polynomial2D"""

        p2 = models.Polynomial2D(5)
        z = p2(self.x, self.y)
        assert z.shape == (10, 10)

    def test_p2_nset_npset(self):
        """N param_sets, N 2D data sets, Poly2d"""

        p2 = models.Polynomial2D(5, n_models=2)
        xx = np.array([self.x, self.x])
        yy = np.array([self.y, self.y])
        z = p2(xx, yy)
        assert z.shape == (2, 10, 10)

    def test_nset_domain(self):
        """
        Test model set with negative model_set_axis.

        In this case model_set_axis=-1 is identical to model_set_axis=1.
        """

        xx = np.array([self.x1, self.x1]).T
        xx[0, 0] = 100
        xx[1, 0] = 100
        xx[2, 0] = 99
        p1 = models.Polynomial1D(5, c0=[1, 2], c1=[3, 4], n_models=2)
        yy = p1(xx, model_set_axis=-1)
        assert_allclose(xx.shape, yy.shape)
        yy1 = p1(xx, model_set_axis=1)
        assert_allclose(yy, yy1)

    def test_evaluate_gauss2d(self):
        cov = np.array([[1.0, 0.8], [0.8, 3]])
        g = models.Gaussian2D(1.0, 5.0, 4.0, cov_matrix=cov)
        y, x = np.mgrid[:10, :10]
        g(x, y)


class TModel_1_1(Fittable1DModel):
    p1 = Parameter()
    p2 = Parameter()

    @staticmethod
    def evaluate(x, p1, p2):
        return x + p1 + p2


class TestSingleInputSingleOutputSingleModel:
    """
    A suite of tests to check various cases of parameter and input combinations
    on models with n_input = n_output = 1 on a toy model with n_models=1.

    Many of these tests mirror test cases in
    ``astropy.modeling.tests.test_parameters.TestParameterInitialization``,
    except that this tests how different parameter arrangements interact with
    different types of model inputs.
    """

    def test_scalar_parameters_scalar_input(self):
        """
        Scalar parameters with a scalar input should return a scalar.
        """

        t = TModel_1_1(1, 10)
        y = t(100)
        assert isinstance(y, float)
        assert np.ndim(y) == 0
        assert y == 111

    def test_scalar_parameters_1d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_1(1, 10)
        y = t(np.arange(5) * 100)
        assert isinstance(y, np.ndarray)
        assert np.shape(y) == (5,)
        assert np.all(y == [11, 111, 211, 311, 411])

    def test_scalar_parameters_2d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_1(1, 10)
        y = t(np.arange(6).reshape(2, 3) * 100)
        assert isinstance(y, np.ndarray)
        assert np.shape(y) == (2, 3)
        assert np.all(y == [[11, 111, 211], [311, 411, 511]])

    def test_scalar_parameters_3d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_1(1, 10)
        y = t(np.arange(12).reshape(2, 3, 2) * 100)
        assert isinstance(y, np.ndarray)
        assert np.shape(y) == (2, 3, 2)
        assert np.all(
            y
            == [
                [[11, 111], [211, 311], [411, 511]],
                [[611, 711], [811, 911], [1011, 1111]],
            ]
        )

    def test_1d_array_parameters_scalar_input(self):
        """
        Array parameters should all be broadcastable with each other, and with
        a scalar input the output should be broadcast to the maximum dimensions
        of the parameters.
        """

        t = TModel_1_1([1, 2], [10, 20])
        y = t(100)
        assert isinstance(y, np.ndarray)
        assert np.shape(y) == (2,)
        assert np.all(y == [111, 122])

    def test_1d_array_parameters_1d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_1([1, 2], [10, 20])
        y1 = t([100, 200])
        assert np.shape(y1) == (2,)
        assert np.all(y1 == [111, 222])

        y2 = t([[100], [200]])
        assert np.shape(y2) == (2, 2)
        assert np.all(y2 == [[111, 122], [211, 222]])

        with pytest.raises(ValueError, match="broadcast"):
            # Doesn't broadcast
            t([100, 200, 300])

    def test_2d_array_parameters_2d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_1([[1, 2], [3, 4]], [[10, 20], [30, 40]])

        y1 = t([[100, 200], [300, 400]])
        assert np.shape(y1) == (2, 2)
        assert np.all(y1 == [[111, 222], [333, 444]])

        y2 = t([[[[100]], [[200]]], [[[300]], [[400]]]])
        assert np.shape(y2) == (2, 2, 2, 2)
        assert np.all(
            y2
            == [
                [[[111, 122], [133, 144]], [[211, 222], [233, 244]]],
                [[[311, 322], [333, 344]], [[411, 422], [433, 444]]],
            ]
        )

        with pytest.raises(ValueError, match="broadcast"):
            # Doesn't broadcast
            t([[100, 200, 300], [400, 500, 600]])

    def test_mixed_array_parameters_1d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_1(
            [
                [[0.01, 0.02, 0.03], [0.04, 0.05, 0.06]],
                [[0.07, 0.08, 0.09], [0.10, 0.11, 0.12]],
            ],
            [1, 2, 3],
        )

        y1 = t([10, 20, 30])
        assert np.shape(y1) == (2, 2, 3)
        assert_allclose(
            y1,
            [
                [[11.01, 22.02, 33.03], [11.04, 22.05, 33.06]],
                [[11.07, 22.08, 33.09], [11.10, 22.11, 33.12]],
            ],
        )

        y2 = t([[[[10]]], [[[20]]], [[[30]]]])
        assert np.shape(y2) == (3, 2, 2, 3)
        assert_allclose(
            y2,
            [
                [
                    [[11.01, 12.02, 13.03], [11.04, 12.05, 13.06]],
                    [[11.07, 12.08, 13.09], [11.10, 12.11, 13.12]],
                ],
                [
                    [[21.01, 22.02, 23.03], [21.04, 22.05, 23.06]],
                    [[21.07, 22.08, 23.09], [21.10, 22.11, 23.12]],
                ],
                [
                    [[31.01, 32.02, 33.03], [31.04, 32.05, 33.06]],
                    [[31.07, 32.08, 33.09], [31.10, 32.11, 33.12]],
                ],
            ],
        )


class TestSingleInputSingleOutputTwoModel:
    """
    A suite of tests to check various cases of parameter and input combinations
    on models with n_input = n_output = 1 on a toy model with n_models=2.

    Many of these tests mirror test cases in
    ``astropy.modeling.tests.test_parameters.TestParameterInitialization``,
    except that this tests how different parameter arrangements interact with
    different types of model inputs.

    With n_models=2 all outputs should have a first dimension of size 2 (unless
    defined with model_set_axis != 0).
    """

    def test_scalar_parameters_scalar_input(self):
        """
        Scalar parameters with a scalar input should return a 1-D array with
        size equal to the number of models.
        """

        t = TModel_1_1([1, 2], [10, 20], n_models=2)

        y = t(100)
        assert np.shape(y) == (2,)
        assert np.all(y == [111, 122])

    def test_scalar_parameters_1d_array_input(self):
        """
        The dimension of the input should match the number of models unless
        model_set_axis=False is given, in which case the input is copied across
        all models.
        """

        t = TModel_1_1([1, 2], [10, 20], n_models=2)

        MESSAGE = (
            r"Input argument .* does not have the correct dimensions in .* for a model"
            r" set with .*"
        )
        with pytest.raises(ValueError, match=MESSAGE):
            t(np.arange(5) * 100)

        y1 = t([100, 200])
        assert np.shape(y1) == (2,)
        assert np.all(y1 == [111, 222])

        y2 = t([100, 200], model_set_axis=False)
        # In this case the value [100, 200, 300] should be evaluated on each
        # model rather than evaluating the first model with 100 and the second
        # model  with 200
        assert np.shape(y2) == (2, 2)
        assert np.all(y2 == [[111, 211], [122, 222]])

        y3 = t([100, 200, 300], model_set_axis=False)
        assert np.shape(y3) == (2, 3)
        assert np.all(y3 == [[111, 211, 311], [122, 222, 322]])

    def test_scalar_parameters_2d_array_input(self):
        """
        The dimension of the input should match the number of models unless
        model_set_axis=False is given, in which case the input is copied across
        all models.
        """

        t = TModel_1_1([1, 2], [10, 20], n_models=2)

        y1 = t(np.arange(6).reshape(2, 3) * 100)
        assert np.shape(y1) == (2, 3)
        assert np.all(y1 == [[11, 111, 211], [322, 422, 522]])

        y2 = t(np.arange(6).reshape(2, 3) * 100, model_set_axis=False)
        assert np.shape(y2) == (2, 2, 3)
        assert np.all(
            y2 == [[[11, 111, 211], [311, 411, 511]], [[22, 122, 222], [322, 422, 522]]]
        )

    def test_scalar_parameters_3d_array_input(self):
        """
        The dimension of the input should match the number of models unless
        model_set_axis=False is given, in which case the input is copied across
        all models.
        """

        t = TModel_1_1([1, 2], [10, 20], n_models=2)
        data = np.arange(12).reshape(2, 3, 2) * 100

        y1 = t(data)
        assert np.shape(y1) == (2, 3, 2)
        assert np.all(
            y1
            == [
                [[11, 111], [211, 311], [411, 511]],
                [[622, 722], [822, 922], [1022, 1122]],
            ]
        )

        y2 = t(data, model_set_axis=False)
        assert np.shape(y2) == (2, 2, 3, 2)
        assert np.all(y2 == np.array([data + 11, data + 22]))

    def test_1d_array_parameters_scalar_input(self):
        """
        Array parameters should all be broadcastable with each other, and with
        a scalar input the output should be broadcast to the maximum dimensions
        of the parameters.
        """

        t = TModel_1_1([[1, 2, 3], [4, 5, 6]], [[10, 20, 30], [40, 50, 60]], n_models=2)

        y = t(100)
        assert np.shape(y) == (2, 3)
        assert np.all(y == [[111, 122, 133], [144, 155, 166]])

    def test_1d_array_parameters_1d_array_input(self):
        """
        When the input is an array, if model_set_axis=False then it must
        broadcast with the shapes of the parameters (excluding the
        model_set_axis).

        Otherwise all dimensions must be broadcastable.
        """

        t = TModel_1_1([[1, 2, 3], [4, 5, 6]], [[10, 20, 30], [40, 50, 60]], n_models=2)

        MESSAGE = (
            r"Input argument .* does not have the correct dimensions in .* for a model"
            r" set with .*"
        )
        with pytest.raises(ValueError, match=MESSAGE):
            y1 = t([100, 200, 300])

        y1 = t([100, 200])
        assert np.shape(y1) == (2, 3)
        assert np.all(y1 == [[111, 122, 133], [244, 255, 266]])

        with pytest.raises(ValueError, match="broadcast"):
            # Doesn't broadcast with the shape of the parameters, (3,)
            y2 = t([100, 200], model_set_axis=False)

        y2 = t([100, 200, 300], model_set_axis=False)
        assert np.shape(y2) == (2, 3)
        assert np.all(y2 == [[111, 222, 333], [144, 255, 366]])

    def test_2d_array_parameters_2d_array_input(self):
        t = TModel_1_1(
            [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
            [[[10, 20], [30, 40]], [[50, 60], [70, 80]]],
            n_models=2,
        )
        y1 = t([[100, 200], [300, 400]])
        assert np.shape(y1) == (2, 2, 2)
        assert np.all(
            y1
            == [
                [[111, 222], [133, 244]],
                [[355, 466], [377, 488]],
            ]
        )

        with pytest.raises(ValueError, match="broadcast"):
            y2 = t([[100, 200, 300], [400, 500, 600]])

        y2 = t([[[100, 200], [300, 400]], [[500, 600], [700, 800]]])
        assert np.shape(y2) == (2, 2, 2)
        assert np.all(
            y2
            == [
                [[111, 222], [333, 444]],
                [[555, 666], [777, 888]],
            ]
        )

    def test_mixed_array_parameters_1d_array_input(self):
        t = TModel_1_1(
            [
                [[0.01, 0.02, 0.03], [0.04, 0.05, 0.06]],
                [[0.07, 0.08, 0.09], [0.10, 0.11, 0.12]],
            ],
            [[1, 2, 3], [4, 5, 6]],
            n_models=2,
        )

        MESSAGE = (
            r"Input argument .* does not have the correct dimensions in .* for a model"
            r" set with .*"
        )
        with pytest.raises(ValueError, match=MESSAGE):
            y = t([10, 20, 30])

        y = t([10, 20, 30], model_set_axis=False)
        assert np.shape(y) == (2, 2, 3)
        assert_allclose(
            y,
            [
                [[11.01, 22.02, 33.03], [11.04, 22.05, 33.06]],
                [[14.07, 25.08, 36.09], [14.10, 25.11, 36.12]],
            ],
        )


class TModel_1_2(FittableModel):
    n_inputs = 1
    n_outputs = 2

    p1 = Parameter()
    p2 = Parameter()
    p3 = Parameter()

    @staticmethod
    def evaluate(x, p1, p2, p3):
        return (x + p1 + p2, x + p1 + p2 + p3)


class TestSingleInputDoubleOutputSingleModel:
    """
    A suite of tests to check various cases of parameter and input combinations
    on models with n_input = 1 but n_output = 2 on a toy model with n_models=1.

    As of writing there are not enough controls to adjust how outputs from such
    a model should be formatted (currently the shapes of outputs are assumed to
    be directly associated with the shapes of corresponding inputs when
    n_inputs == n_outputs).  For now, the approach taken for cases like this is
    to assume all outputs should have the same format.
    """

    def test_scalar_parameters_scalar_input(self):
        """
        Scalar parameters with a scalar input should return a scalar.
        """

        t = TModel_1_2(1, 10, 1000)
        y, z = t(100)
        assert isinstance(y, float)
        assert isinstance(z, float)
        assert np.ndim(y) == np.ndim(z) == 0
        assert y == 111
        assert z == 1111

    def test_scalar_parameters_1d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_2(1, 10, 1000)
        y, z = t(np.arange(5) * 100)
        assert isinstance(y, np.ndarray)
        assert isinstance(z, np.ndarray)
        assert np.shape(y) == np.shape(z) == (5,)
        assert np.all(y == [11, 111, 211, 311, 411])
        assert np.all(z == (y + 1000))

    def test_scalar_parameters_2d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_2(1, 10, 1000)
        y, z = t(np.arange(6).reshape(2, 3) * 100)
        assert isinstance(y, np.ndarray)
        assert isinstance(z, np.ndarray)
        assert np.shape(y) == np.shape(z) == (2, 3)
        assert np.all(y == [[11, 111, 211], [311, 411, 511]])
        assert np.all(z == (y + 1000))

    def test_scalar_parameters_3d_array_input(self):
        """
        Scalar parameters should broadcast with an array input to result in an
        array output of the same shape as the input.
        """

        t = TModel_1_2(1, 10, 1000)
        y, z = t(np.arange(12).reshape(2, 3, 2) * 100)
        assert isinstance(y, np.ndarray)
        assert isinstance(z, np.ndarray)
        assert np.shape(y) == np.shape(z) == (2, 3, 2)
        assert np.all(
            y
            == [
                [[11, 111], [211, 311], [411, 511]],
                [[611, 711], [811, 911], [1011, 1111]],
            ]
        )
        assert np.all(z == (y + 1000))

    def test_1d_array_parameters_scalar_input(self):
        """
        Array parameters should all be broadcastable with each other, and with
        a scalar input the output should be broadcast to the maximum dimensions
        of the parameters.
        """

        t = TModel_1_2([1, 2], [10, 20], [1000, 2000])
        y, z = t(100)
        assert isinstance(y, np.ndarray)
        assert isinstance(z, np.ndarray)
        assert np.shape(y) == np.shape(z) == (2,)
        assert np.all(y == [111, 122])
        assert np.all(z == [1111, 2122])

    def test_1d_array_parameters_1d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_2([1, 2], [10, 20], [1000, 2000])
        y1, z1 = t([100, 200])
        assert np.shape(y1) == np.shape(z1) == (2,)
        assert np.all(y1 == [111, 222])
        assert np.all(z1 == [1111, 2222])

        y2, z2 = t([[100], [200]])
        assert np.shape(y2) == np.shape(z2) == (2, 2)
        assert np.all(y2 == [[111, 122], [211, 222]])
        assert np.all(z2 == [[1111, 2122], [1211, 2222]])

        with pytest.raises(ValueError, match="broadcast"):
            # Doesn't broadcast
            y3, z3 = t([100, 200, 300])

    def test_2d_array_parameters_2d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_2(
            [[1, 2], [3, 4]], [[10, 20], [30, 40]], [[1000, 2000], [3000, 4000]]
        )

        y1, z1 = t([[100, 200], [300, 400]])
        assert np.shape(y1) == np.shape(z1) == (2, 2)
        assert np.all(y1 == [[111, 222], [333, 444]])
        assert np.all(z1 == [[1111, 2222], [3333, 4444]])

        y2, z2 = t([[[[100]], [[200]]], [[[300]], [[400]]]])
        assert np.shape(y2) == np.shape(z2) == (2, 2, 2, 2)
        assert np.all(
            y2
            == [
                [[[111, 122], [133, 144]], [[211, 222], [233, 244]]],
                [[[311, 322], [333, 344]], [[411, 422], [433, 444]]],
            ]
        )
        assert np.all(
            z2
            == [
                [[[1111, 2122], [3133, 4144]], [[1211, 2222], [3233, 4244]]],
                [[[1311, 2322], [3333, 4344]], [[1411, 2422], [3433, 4444]]],
            ]
        )

        with pytest.raises(ValueError, match="broadcast"):
            # Doesn't broadcast
            y3, z3 = t([[100, 200, 300], [400, 500, 600]])

    def test_mixed_array_parameters_1d_array_input(self):
        """
        When given an array input it must be broadcastable with all the
        parameters.
        """

        t = TModel_1_2(
            [
                [[0.01, 0.02, 0.03], [0.04, 0.05, 0.06]],
                [[0.07, 0.08, 0.09], [0.10, 0.11, 0.12]],
            ],
            [1, 2, 3],
            [100, 200, 300],
        )

        y1, z1 = t([10, 20, 30])
        assert np.shape(y1) == np.shape(z1) == (2, 2, 3)
        assert_allclose(
            y1,
            [
                [[11.01, 22.02, 33.03], [11.04, 22.05, 33.06]],
                [[11.07, 22.08, 33.09], [11.10, 22.11, 33.12]],
            ],
        )
        assert_allclose(
            z1,
            [
                [[111.01, 222.02, 333.03], [111.04, 222.05, 333.06]],
                [[111.07, 222.08, 333.09], [111.10, 222.11, 333.12]],
            ],
        )

        y2, z2 = t([[[[10]]], [[[20]]], [[[30]]]])
        assert np.shape(y2) == np.shape(z2) == (3, 2, 2, 3)
        assert_allclose(
            y2,
            [
                [
                    [[11.01, 12.02, 13.03], [11.04, 12.05, 13.06]],
                    [[11.07, 12.08, 13.09], [11.10, 12.11, 13.12]],
                ],
                [
                    [[21.01, 22.02, 23.03], [21.04, 22.05, 23.06]],
                    [[21.07, 22.08, 23.09], [21.10, 22.11, 23.12]],
                ],
                [
                    [[31.01, 32.02, 33.03], [31.04, 32.05, 33.06]],
                    [[31.07, 32.08, 33.09], [31.10, 32.11, 33.12]],
                ],
            ],
        )

        assert_allclose(
            z2,
            [
                [
                    [[111.01, 212.02, 313.03], [111.04, 212.05, 313.06]],
                    [[111.07, 212.08, 313.09], [111.10, 212.11, 313.12]],
                ],
                [
                    [[121.01, 222.02, 323.03], [121.04, 222.05, 323.06]],
                    [[121.07, 222.08, 323.09], [121.10, 222.11, 323.12]],
                ],
                [
                    [[131.01, 232.02, 333.03], [131.04, 232.05, 333.06]],
                    [[131.07, 232.08, 333.09], [131.10, 232.11, 333.12]],
                ],
            ],
        )


# test broadcasting rules
broadcast_models = [
    {"model": models.Identity(2), "inputs": [0, [1, 1]], "outputs": [0, [1, 1]]},
    {"model": models.Identity(2), "inputs": [[1, 1], 0], "outputs": [[1, 1], 0]},
    {"model": models.Mapping((0, 1)), "inputs": [0, [1, 1]], "outputs": [0, [1, 1]]},
    {"model": models.Mapping((1, 0)), "inputs": [0, [1, 1]], "outputs": [[1, 1], 0]},
    {
        "model": models.Mapping((1, 0), n_inputs=3),
        "inputs": [0, [1, 1], 2],
        "outputs": [[1, 1], 0],
    },
    {
        "model": models.Mapping((0, 1, 0)),
        "inputs": [0, [1, 1]],
        "outputs": [0, [1, 1], 0],
    },
    {
        "model": models.Mapping((0, 1, 1)),
        "inputs": [0, [1, 1]],
        "outputs": [0, [1, 1], [1, 1]],
    },
    {"model": models.Polynomial2D(1, c0_0=1), "inputs": [0, [1, 1]], "outputs": [1, 1]},
    {"model": models.Polynomial2D(1, c0_0=1), "inputs": [0, 1], "outputs": 1},
    {
        "model": models.Gaussian2D(1, 1, 2, 1, 1.2),
        "inputs": [0, [1, 1]],
        "outputs": [0.42860385, 0.42860385],
    },
    {
        "model": models.Gaussian2D(1, 1, 2, 1, 1.2),
        "inputs": [0, 1],
        "outputs": 0.428603846153,
    },
    {
        "model": models.Polynomial2D(1, c0_0=1) & models.Polynomial2D(1, c0_0=2),
        "inputs": [1, 1, 1, 1],
        "outputs": (1, 2),
    },
    {
        "model": models.Polynomial2D(1, c0_0=1) & models.Polynomial2D(1, c0_0=2),
        "inputs": [1, 1, [1, 1], [1, 1]],
        "outputs": (1, [2, 2]),
    },
    {
        "model": models.math.MultiplyUfunc(),
        "inputs": [np.array([np.linspace(0, 1, 5)]).T, np.arange(2)],
        "outputs": np.array(
            [[0.0, 0.0], [0.0, 0.25], [0.0, 0.5], [0.0, 0.75], [0.0, 1.0]]
        ),
    },
]


@pytest.mark.parametrize("model", broadcast_models)
def test_mixed_input(model):
    result = model["model"](*model["inputs"])
    if np.isscalar(result):
        assert_allclose(result, model["outputs"])
    else:
        for i in range(len(result)):
            assert_allclose(result[i], model["outputs"][i])


def test_more_outputs():
    class M(FittableModel):
        standard_broadcasting = False
        n_inputs = 2
        n_outputs = 3

        a = Parameter()

        def evaluate(self, x, y, a):
            return a * x, a - x, a + y

        def __call__(self, *args, **kwargs):
            inputs, _ = super().prepare_inputs(*args, **kwargs)

            outputs = self.evaluate(*inputs, *self.parameters)
            output_shapes = [out.shape for out in outputs]
            output_shapes = [() if shape == (1,) else shape for shape in output_shapes]
            return self.prepare_outputs((tuple(output_shapes),), *outputs, **kwargs)

    c = M(1)
    result = c([1, 1], 1)
    expected = [[1.0, 1.0], [0.0, 0.0], 2.0]
    for r, e in zip(result, expected):
        assert_allclose(r, e)

    c = M(1)
    result = c(1, [1, 1])
    expected = [1.0, 0.0, [2.0, 2.0]]
    for r, e in zip(result, expected):
        assert_allclose(r, e)


class TInputFormatter(Model):
    """
    A toy model to test input/output formatting.
    """

    n_inputs = 2
    n_outputs = 2
    _outputs = ("x", "y")

    @staticmethod
    def evaluate(x, y):
        return x, y


def test_format_input_scalars():
    model = TInputFormatter()
    result = model(1, 2)
    assert result == (1, 2)


def test_format_input_arrays():
    model = TInputFormatter()
    result = model([1, 1], [2, 2])
    assert_allclose(result, (np.array([1, 1]), np.array([2, 2])))


def test_format_input_arrays_transposed():
    model = TInputFormatter()
    input = np.array([[1, 1]]).T, np.array([[2, 2]]).T
    result = model(*input)
    assert_allclose(result, input)


@pytest.mark.parametrize(
    "cls, kwargs",
    [
        pytest.param(models.Gaussian2D, {}, id="Gaussian2D"),
        pytest.param(models.Polynomial2D, {"degree": 1}, id="Polynomial2D"),
        pytest.param(models.Rotation2D, {}, id="Rotation2D"),
        pytest.param(models.Pix2Sky_TAN, {}, id="Pix2Sky_TAN"),
        pytest.param(
            models.Tabular2D, {"lookup_table": np.ones((4, 5))}, id="Tabular2D"
        ),
    ],
)
@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_call_keyword_args_1(cls, kwargs):
    """
    Test calling a model with positional, keywrd and a mixture of both arguments.
    """
    model = cls(**kwargs)
    positional = model(1, 2)
    assert_allclose(positional, model(x=1, y=2))
    assert_allclose(positional, model(1, y=2))

    model.inputs = ("r", "t")
    assert_allclose(positional, model(r=1, t=2))
    assert_allclose(positional, model(1, t=2))
    assert_allclose(positional, model(1, 2))

    MESSAGE = r"Too many input arguments - expected 2, got .*"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, 3)

    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, t=12, r=3)

    MESSAGE = r"Missing input arguments - expected 2, got 1"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1)


@pytest.mark.parametrize(
    "cls, kwargs",
    [
        pytest.param(models.Gaussian1D, {}, id="Gaussian1D"),
        pytest.param(models.Polynomial1D, {"degree": 1}, id="Polynomial1D"),
        pytest.param(models.Tabular1D, {"lookup_table": np.ones((5,))}, id="Tabular1D"),
    ],
)
@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_call_keyword_args_2(cls, kwargs):
    """
    Test calling a model with positional, keywrd and a mixture of both arguments.
    """
    model = cls(**kwargs)
    positional = model(1)
    assert_allclose(positional, model(x=1))

    model.inputs = ("r",)
    assert_allclose(positional, model(r=1))

    MESSAGE = r"Too many input arguments - expected .*, got .*"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, 3)

    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, t=12, r=3)

    MESSAGE = r"Missing input arguments - expected 1, got 0"
    with pytest.raises(ValueError, match=MESSAGE):
        model()


@pytest.mark.parametrize(
    "model",
    [
        models.Gaussian2D() | models.Polynomial1D(1),
        models.Gaussian1D() & models.Polynomial1D(1),
        models.Gaussian2D() + models.Polynomial2D(1),
        models.Gaussian2D() - models.Polynomial2D(1),
        models.Gaussian2D() * models.Polynomial2D(1),
        models.Identity(2) | models.Polynomial2D(1),
        models.Mapping((1,)) | models.Polynomial1D(1),
    ],
)
def test_call_keyword_args_3(model):
    """
    Test calling a model with positional, keywrd and a mixture of both arguments.
    """
    positional = model(1, 2)
    model.inputs = ("r", "t")
    assert_allclose(positional, model(r=1, t=2))

    assert_allclose(positional, model(1, t=2))

    MESSAGE = r"Too many input arguments - expected .*, got .*"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, 3)

    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, t=12, r=3)

    MESSAGE = r"Missing input arguments - expected 2, got 0"
    with pytest.raises(ValueError, match=MESSAGE):
        model()


@pytest.mark.parametrize(
    "model",
    [
        models.Identity(2),
        models.Mapping((0, 1)),
        models.Mapping((1,)),
    ],
)
def test_call_keyword_mappings(model):
    """
    Test calling a model with positional, keywrd and a mixture of both arguments.
    """
    positional = model(1, 2)
    assert_allclose(positional, model(x0=1, x1=2))
    assert_allclose(positional, model(1, x1=2))

    # We take a copy before modifying the model since otherwise this changes
    # the instance used in the parametrize call and affects future test runs.
    model = model.copy()

    model.inputs = ("r", "t")
    assert_allclose(positional, model(r=1, t=2))
    assert_allclose(positional, model(1, t=2))
    assert_allclose(positional, model(1, 2))

    MESSAGE = r"Too many input arguments - expected .*, got .*"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, 3)

    with pytest.raises(ValueError, match=MESSAGE):
        model(1, 2, t=12, r=3)

    MESSAGE = r"Missing input arguments - expected 2, got 1"
    with pytest.raises(ValueError, match=MESSAGE):
        model(1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""The class that implements the framework for defining graphs of transformations.

The fundamental idea is that each class is a node in the transformation graph,
and transitions from one node to another are defined as functions (or methods)
wrapped in transformation objects.
"""

import functools
import heapq
import subprocess
from collections import defaultdict
from contextlib import contextmanager
from itertools import pairwise

from astropy.coordinates.transformations.affine import (
    AffineTransform,
    DynamicMatrixTransform,
    StaticMatrixTransform,
)
from astropy.coordinates.transformations.composite import CompositeTransform
from astropy.coordinates.transformations.function import (
    FunctionTransform,
    FunctionTransformWithFiniteDifference,
)

__all__ = ["TransformGraph"]


# map class names to colorblind-safe colors
trans_to_color = {}
trans_to_color[AffineTransform] = "#555555"  # gray
trans_to_color[FunctionTransform] = "#783001"  # dark red-ish/brown
trans_to_color[FunctionTransformWithFiniteDifference] = "#d95f02"  # red-ish
trans_to_color[StaticMatrixTransform] = "#7570b3"  # blue-ish
trans_to_color[DynamicMatrixTransform] = "#1b9e77"  # green-ish


def frame_attrs_from_set(frame_set):
    """
    A `dict` of all the attributes of all frame classes in this
    `~astropy.coordinates.TransformGraph`.

    Broken out of the class so this can be called on a temporary frame set to
    validate new additions to the transform graph before actually adding them.
    """
    result = {}

    for frame_cls in frame_set:
        result.update(frame_cls.frame_attributes)

    return result


def frame_comps_from_set(frame_set):
    """
    A `set` of all component names every defined within any frame class in
    this `~astropy.coordinates.TransformGraph`.

    Broken out of the class so this can be called on a temporary frame set to
    validate new additions to the transform graph before actually adding them.
    """
    result = set()

    for frame_cls in frame_set:
        rep_info = frame_cls._frame_specific_representation_info
        for mappings in rep_info.values():
            for rep_map in mappings:
                result.update([rep_map.framename])

    return result


class TransformGraph:
    """
    A graph representing the paths between coordinate frames.
    """

    def __init__(self):
        self._graph = defaultdict(dict)
        self.invalidate_cache()  # generates cache entries

    @functools.cached_property
    def _cached_names(self):
        dct = {}
        for c in self.frame_set:
            if (nm := getattr(c, "name", None)) is not None:
                dct |= dict.fromkeys(nm if isinstance(nm, list) else [nm], c)
        return dct

    @property
    def frame_set(self):
        """
        A `set` of all the frame classes present in this TransformGraph.
        """
        if self._cached_frame_set is None:
            self._cached_frame_set = set()
            for a in self._graph:
                self._cached_frame_set.add(a)
                for b in self._graph[a]:
                    self._cached_frame_set.add(b)

        return self._cached_frame_set.copy()

    @functools.cached_property
    def frame_attributes(self):
        """
        A `dict` of all the attributes of all frame classes in this TransformGraph.
        """
        return frame_attrs_from_set(self.frame_set)

    @property
    def frame_component_names(self):
        """
        A `set` of all component names every defined within any frame class in
        this TransformGraph.
        """
        return frame_comps_from_set(self.frame_set)

    def invalidate_cache(self):
        """
        Invalidates the cache that stores optimizations for traversing the
        transform graph.  This is called automatically when transforms
        are added or removed, but will need to be called manually if
        weights on transforms are modified inplace.
        """
        vars(self).pop("_cached_names", None)
        self._cached_frame_set = None
        vars(self).pop("frame_attributes", None)
        self._shortestpaths = {}
        self._composite_cache = {}

    def add_transform(self, fromsys, tosys, transform):
        """Add a new coordinate transformation to the graph.

        Parameters
        ----------
        fromsys : class
            The coordinate frame class to start from.
        tosys : class
            The coordinate frame class to transform into.
        transform : `~astropy.coordinates.CoordinateTransform`
            The transformation object. Typically a
            `~astropy.coordinates.CoordinateTransform` object, although it may
            be some other callable that is called with the same signature.

        Raises
        ------
        TypeError
            If ``fromsys`` or ``tosys`` are not classes or ``transform`` is
            not callable.

        """
        if not isinstance(fromsys, type):
            raise TypeError("fromsys must be a class")
        if not isinstance(tosys, type):
            raise TypeError("tosys must be a class")
        if not callable(transform):
            raise TypeError("transform must be callable")

        frame_set = self.frame_set.copy()
        frame_set.add(fromsys)
        frame_set.add(tosys)

        # Now we check to see if any attributes on the proposed frames override
        # *any* component names, which we can't allow for some of the logic in
        # the SkyCoord initializer to work
        attrs = set(frame_attrs_from_set(frame_set).keys())
        comps = frame_comps_from_set(frame_set)

        invalid_attrs = attrs.intersection(comps)
        if invalid_attrs:
            invalid_frames = set()
            for attr in invalid_attrs:
                if attr in fromsys.frame_attributes:
                    invalid_frames.update([fromsys])

                if attr in tosys.frame_attributes:
                    invalid_frames.update([tosys])

            raise ValueError(
                f"Frame(s) {list(invalid_frames)} contain invalid attribute names:"
                f" {invalid_attrs}\nFrame attributes can not conflict with *any* of"
                " the frame data component names (see"
                " `frame_transform_graph.frame_component_names`)."
            )

        self._graph[fromsys][tosys] = transform
        self.invalidate_cache()

    def remove_transform(self, fromsys, tosys, transform):
        """
        Removes a coordinate transform from the graph.

        Parameters
        ----------
        fromsys : class or None
            The coordinate frame *class* to start from. If `None`,
            ``transform`` will be searched for and removed (``tosys`` must
            also be `None`).
        tosys : class or None
            The coordinate frame *class* to transform into. If `None`,
            ``transform`` will be searched for and removed (``fromsys`` must
            also be `None`).
        transform : callable or None
            The transformation object to be removed or `None`.  If `None`
            and ``tosys`` and ``fromsys`` are supplied, there will be no
            check to ensure the correct object is removed.
        """
        if fromsys is None or tosys is None:
            if tosys is not fromsys:
                raise ValueError("fromsys and tosys must both be None if either are")
            if transform is None:
                raise ValueError("cannot give all Nones to remove_transform")

            # search for the requested transform by brute force and remove it
            for a, agraph in self._graph.items():
                for b, bgraph in agraph.items():
                    if bgraph is transform:
                        del agraph[b]
                        fromsys = a
                        break

                # If the transform was found, need to break out of the outer for loop too
                if fromsys:
                    break
            else:
                raise ValueError(f"Could not find transform {transform} in the graph")

        elif transform is None or self._graph[fromsys].get(tosys, None) is transform:
            self._graph[fromsys].pop(tosys, None)
        else:
            raise ValueError(
                f"Current transform from {fromsys} to {tosys} is not {transform}"
            )

        # Remove the subgraph if it is now empty
        if self._graph[fromsys] == {}:
            self._graph.pop(fromsys)

        self.invalidate_cache()

    def find_shortest_path(self, fromsys, tosys):
        """
        Computes the shortest distance along the transform graph from
        one system to another.

        Parameters
        ----------
        fromsys : class
            The coordinate frame class to start from.
        tosys : class
            The coordinate frame class to transform into.

        Returns
        -------
        path : list of class or None
            The path from ``fromsys`` to ``tosys`` as an in-order sequence
            of classes.  This list includes *both* ``fromsys`` and
            ``tosys``. Is `None` if there is no possible path.
        distance : float or int
            The total distance/priority from ``fromsys`` to ``tosys``.  If
            priorities are not set this is the number of transforms
            needed. Is ``inf`` if there is no possible path.
        """
        inf = float("inf")

        # special-case the 0 or 1-path
        if tosys is fromsys and tosys not in self._graph[fromsys]:
            # Means there's no transform necessary to go from it to itself.
            return [tosys], 0
        if tosys in self._graph[fromsys]:
            # this will also catch the case where tosys is fromsys, but has
            # a defined transform.
            t = self._graph[fromsys][tosys]
            return [fromsys, tosys], float(getattr(t, "priority", 1))

        # otherwise, need to construct the path:

        if fromsys in self._shortestpaths:
            # already have a cached result
            return self._shortestpaths[fromsys].get(tosys, (None, inf))

        # use Dijkstra's algorithm to find shortest path in all other cases

        # We store nodes as `dict` keys because differently from `list` uniqueness is
        # guaranteed and differently from `set` insertion order is preserved.
        nodes = {}
        for node, node_graph in self._graph.items():
            nodes[node] = None
            nodes |= dict.fromkeys(node_graph)

        if fromsys not in nodes or tosys not in nodes:
            # fromsys or tosys are isolated or not registered, so there's
            # certainly no way to get from one to the other
            return None, inf

        edgeweights = {}
        # construct another graph that is a dict of dicts of priorities
        # (used as edge weights in Dijkstra's algorithm)
        for a, graph in self._graph.items():
            edgeweights[a] = {b: float(getattr(graph[b], "priority", 1)) for b in graph}

        # entries in q are [distance, count, nodeobj, pathlist]
        # count is needed because in py 3.x, tie-breaking fails on the nodes.
        # this way, insertion order is preserved if the weights are the same
        q = [[0, -1, fromsys, []]]
        q.extend([inf, i, n, []] for i, n in enumerate(nodes) if n is not fromsys)

        # this dict will store the distance to node from ``fromsys`` and the path
        result = {}

        # definitely starts as a valid heap because of the insert line; from the
        # node to itself is always the shortest distance
        while q:
            d, orderi, n, path = heapq.heappop(q)

            if d == inf:
                # everything left is unreachable from fromsys, just copy them to
                # the results and jump out of the loop
                result[n] = (None, d)
                for d, _, n, _ in q:
                    result[n] = (None, d)
                break
            result[n] = (path, d)
            path.append(n)
            if n not in edgeweights:
                # this is a system that can be transformed to, but not from.
                continue
            for n2 in edgeweights[n]:
                if n2 not in result:  # already visited
                    # find where n2 is in the heap
                    for q_elem in q:
                        if q_elem[2] == n2:
                            if (newd := d + edgeweights[n][n2]) < q_elem[0]:
                                q_elem[0] = newd
                                q_elem[3] = list(path)
                                heapq.heapify(q)
                            break
                    else:
                        raise ValueError("n2 not in heap - this should be impossible!")

        # cache for later use
        self._shortestpaths[fromsys] = result
        return result[tosys]

    def get_transform(self, fromsys, tosys):
        """Generates and returns the CompositeTransform for a transformation
        between two coordinate systems.

        Parameters
        ----------
        fromsys : class
            The coordinate frame class to start from.
        tosys : class
            The coordinate frame class to transform into.

        Returns
        -------
        trans : `~astropy.coordinates.CompositeTransform` or None
            If there is a path from ``fromsys`` to ``tosys``, this is a
            transform object for that path.   If no path could be found, this is
            `None`.

        Notes
        -----
        A `~astropy.coordinates.CompositeTransform` is always returned, because
        `~astropy.coordinates.CompositeTransform` is slightly more adaptable in
        the way it can be called than other transform classes. Specifically, it
        takes care of intermediate steps of transformations in a way that is
        consistent with 1-hop transformations.

        """
        if not isinstance(fromsys, type):
            raise TypeError("fromsys is not a class")
        if not isinstance(tosys, type):
            raise TypeError("tosys is not a class")

        path, distance = self.find_shortest_path(fromsys, tosys)

        if path is None:
            return None

        transforms = []
        currsys = fromsys
        for p in path[1:]:  # first element is fromsys so we skip it
            transforms.append(self._graph[currsys][p])
            currsys = p

        fttuple = (fromsys, tosys)
        if fttuple not in self._composite_cache:
            comptrans = CompositeTransform(
                transforms, fromsys, tosys, register_graph=False
            )
            self._composite_cache[fttuple] = comptrans
        return self._composite_cache[fttuple]

    def lookup_name(self, name):
        """
        Tries to locate the coordinate class with the provided alias.

        Parameters
        ----------
        name : str
            The alias to look up.

        Returns
        -------
        `BaseCoordinateFrame` subclass
            The coordinate class corresponding to the ``name`` or `None` if
            no such class exists.
        """
        return self._cached_names.get(name, None)

    def get_names(self):
        """
        Returns all available transform names. They will all be
        valid arguments to `lookup_name`.

        Returns
        -------
        nms : list
            The aliases for coordinate systems.
        """
        return list(self._cached_names.keys())

    def to_dot_graph(
        self,
        priorities=True,
        addnodes=[],
        savefn=None,
        savelayout="plain",
        saveformat=None,
        color_edges=True,
    ):
        """
        Converts this transform graph to the graphviz_ DOT format.

        Optionally saves it (requires `graphviz`_ be installed and on your path).

        .. _graphviz: http://www.graphviz.org/

        Parameters
        ----------
        priorities : bool
            If `True`, show the priority values for each transform.  Otherwise,
            the will not be included in the graph.
        addnodes : sequence of str
            Additional coordinate systems to add (this can include systems
            already in the transform graph, but they will only appear once).
        savefn : None or str
            The file name to save this graph to or `None` to not save
            to a file.
        savelayout : {"plain", "dot", "neato", "fdp", "sfdp", "circo", "twopi", "nop", "nop2", "osage", "patchwork"}
            The graphviz program to use to layout the graph (see
            graphviz_ for details) or 'plain' to just save the DOT graph
            content. Ignored if ``savefn`` is `None`.
        saveformat : str
            The graphviz output format. (e.g. the ``-Txxx`` option for
            the command line program - see graphviz docs for details).
            Ignored if ``savefn`` is `None`.
        color_edges : bool
            Color the edges between two nodes (frames) based on the type of
            transform. ``FunctionTransform``: red, ``StaticMatrixTransform``:
            blue, ``DynamicMatrixTransform``: green.

        Returns
        -------
        dotgraph : str
            A string with the DOT format graph.
        """
        # We store nodes as `dict` keys because differently from `list` uniqueness is
        # guaranteed and differently from `set` insertion order is preserved.
        nodes = {}
        for node, node_graph in self._graph.items():
            nodes[node] = None
            nodes |= dict.fromkeys(node_graph)
        nodes |= dict.fromkeys(addnodes)
        nodenames = []
        invclsaliases = {
            f: [k for k, v in self._cached_names.items() if v == f]
            for f in self.frame_set
        }
        for n in nodes:
            if n in invclsaliases:
                aliases = "`\\n`".join(invclsaliases[n])
                nodenames.append(
                    f'{n.__name__} [shape=oval label="{n.__name__}\\n`{aliases}`"]'
                )
            else:
                nodenames.append(n.__name__ + "[ shape=oval ]")

        edgenames = []
        # Now the edges
        for a, agraph in self._graph.items():
            for b, transform in agraph.items():
                pri = getattr(transform, "priority", 1)
                color = trans_to_color[transform.__class__] if color_edges else "black"
                edgenames.append((a.__name__, b.__name__, pri, color))

        # generate simple dot format graph
        lines = ["digraph AstropyCoordinateTransformGraph {"]
        lines.append("graph [rankdir=LR]")
        lines.append("; ".join(nodenames) + ";")
        for enm1, enm2, weights, color in edgenames:
            labelstr_fmt = "[ {0} {1} ]"
            priority_part = f'label = "{weights}"' if priorities else ""
            color_part = f'color = "{color}"'
            labelstr = labelstr_fmt.format(priority_part, color_part)
            lines.append(f"{enm1} -> {enm2}{labelstr};")
        lines.append("")
        lines.append("overlap=false")
        lines.append("}")
        dotgraph = "\n".join(lines)

        if savefn is not None:
            if savelayout == "plain":
                with open(savefn, "w") as f:
                    f.write(dotgraph)
            # Options from https://graphviz.org/docs/layouts/
            elif savelayout in (
                "dot",
                "neato",
                "fdp",
                "sfdp",
                "circo",
                "twopi",
                "nop",
                "nop2",
                "osage",
                "patchwork",
            ):
                args = [savelayout]
                if saveformat is not None:
                    args.append("-T" + saveformat)
                proc = subprocess.Popen(
                    args,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                stdout, stderr = proc.communicate(dotgraph)
                if proc.returncode != 0:
                    raise OSError("problem running graphviz: \n" + stderr)

                with open(savefn, "w") as f:
                    f.write(stdout)
            else:
                raise NotImplementedError(f'savelayout="{savelayout}" is not supported')

        return dotgraph

    def to_networkx_graph(self):
        """
        Converts this transform graph into a networkx graph.

        .. note::
            You must have the `networkx <https://networkx.github.io/>`_
            package installed for this to work.

        Returns
        -------
        nxgraph : ``networkx.Graph``
            This `~astropy.coordinates.TransformGraph` as a
            `networkx.Graph <https://networkx.github.io/documentation/stable/reference/classes/graph.html>`_.
        """
        import networkx as nx

        nxgraph = nx.Graph()

        # first make the nodes
        for a in self._graph:
            if a not in nxgraph:
                nxgraph.add_node(a)
            for b in self._graph[a]:
                if b not in nxgraph:
                    nxgraph.add_node(b)

        # Now the edges
        for a in self._graph:
            agraph = self._graph[a]
            for b in agraph:
                transform = agraph[b]
                pri = transform.priority if hasattr(transform, "priority") else 1
                color = trans_to_color[transform.__class__]
                nxgraph.add_edge(a, b, weight=pri, color=color)

        return nxgraph

    def transform(self, transcls, fromsys, tosys, priority=1, **kwargs):
        """A function decorator for defining transformations.

        .. note::
            If decorating a static method of a class, ``@staticmethod``
            should be  added *above* this decorator.

        Parameters
        ----------
        transcls : class
            The class of the transformation object to create.
        fromsys : class
            The coordinate frame class to start from.
        tosys : class
            The coordinate frame class to transform into.
        priority : float or int
            The priority if this transform when finding the shortest
            coordinate transform path - large numbers are lower priorities.

        Additional keyword arguments are passed into the ``transcls``
        constructor.

        Returns
        -------
        deco : function
            A function that can be called on another function as a decorator
            (see example).

        Notes
        -----
        This decorator assumes the first argument of the ``transcls``
        initializer accepts a callable, and that the second and third are
        ``fromsys`` and ``tosys``. If this is not true, you should just
        initialize the class manually and use
        `~astropy.coordinates.TransformGraph.add_transform` instead of this
        decorator.

        Examples
        --------
        ::

            graph = TransformGraph()

            class Frame1(BaseCoordinateFrame):
               ...

            class Frame2(BaseCoordinateFrame):
                ...

            @graph.transform(FunctionTransform, Frame1, Frame2)
            def f1_to_f2(f1_obj):
                ... do something with f1_obj ...
                return f2_obj

        """

        def deco(func):
            # this doesn't do anything directly with the transform because
            # ``register_graph=self`` stores it in the transform graph
            # automatically
            transcls(
                func, fromsys, tosys, priority=priority, register_graph=self, **kwargs
            )
            return func

        return deco

    def _add_merged_transform(self, fromsys, tosys, *furthersys, priority=1):
        """
        Add a single-step transform that encapsulates a multi-step transformation path,
        using the transforms that already exist in the graph.

        The created transform internally calls the existing transforms.  If all of the
        transforms are affine, the merged transform is
        `~astropy.coordinates.DynamicMatrixTransform` (if there are no
        origin shifts) or `~astropy.coordinates.AffineTransform`
        (otherwise).  If at least one of the transforms is not affine, the merged
        transform is
        `~astropy.coordinates.FunctionTransformWithFiniteDifference`.

        This method is primarily useful for defining loopback transformations
        (i.e., where ``fromsys`` and the final ``tosys`` are the same).

        Parameters
        ----------
        fromsys : class
            The coordinate frame class to start from.
        tosys : class
            The coordinate frame class to transform to.
        *furthersys : class
            Additional coordinate frame classes to transform to in order.
        priority : number
            The priority of this transform when finding the shortest
            coordinate transform path - large numbers are lower priorities.

        Notes
        -----
        Even though the created transform is a single step in the graph, it
        will still internally call the constituent transforms.  Thus, there is
        no performance benefit for using this created transform.

        For Astropy's built-in frames, loopback transformations typically use
        `~astropy.coordinates.ICRS` to be safe.  Transforming through an inertial
        frame ensures that changes in observation time and observer
        location/velocity are properly accounted for.

        An error will be raised if a direct transform between ``fromsys`` and
        ``tosys`` already exist.
        """
        frames = [fromsys, tosys, *furthersys]
        lastsys = frames[-1]
        full_path = self.get_transform(fromsys, lastsys)
        transforms = [
            self.get_transform(frame_a, frame_b)
            for frame_a, frame_b in pairwise(frames)
        ]
        if None in transforms:
            raise ValueError("This transformation path is not possible")
        if len(full_path.transforms) == 1:
            raise ValueError(
                f"A direct transform for {fromsys.__name__}->{lastsys.__name__} already"
                " exists"
            )

        self.add_transform(
            fromsys,
            lastsys,
            CompositeTransform(
                transforms, fromsys, lastsys, priority=priority
            )._as_single_transform(),
        )

    @contextmanager
    def impose_finite_difference_dt(self, dt):
        """
        Context manager to impose a finite-difference time step on all applicable transformations.

        For each transformation in this transformation graph that has the attribute
        ``finite_difference_dt``, that attribute is set to the provided value.  The only standard
        transformation with this attribute is
        `~astropy.coordinates.FunctionTransformWithFiniteDifference`.

        Parameters
        ----------
        dt : `~astropy.units.Quantity` ['time'] or callable
            If a quantity, this is the size of the differential used to do the finite difference.
            If a callable, should accept ``(fromcoord, toframe)`` and return the ``dt`` value.
        """
        key = "finite_difference_dt"
        saved_settings = []

        try:
            for to_frames in self._graph.values():
                for transform in to_frames.values():
                    if hasattr(transform, key):
                        old_setting = (transform, key, getattr(transform, key))
                        saved_settings.append(old_setting)
                        setattr(transform, key, dt)
            yield
        finally:
            for setting in saved_settings:
                setattr(*setting)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Parameter test mixin classes."""

import copy
from inspect import BoundArguments

import numpy as np
import pytest

import astropy.constants as const
import astropy.units as u
from astropy.cosmology import Cosmology, FlatFLRWMixin, Parameter
from astropy.cosmology._src.parameter import MISSING
from astropy.cosmology._src.tests.test_core import ParameterTestMixin
from astropy.tests.helper import assert_quantity_allclose


class ParameterH0TestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` H0 on a Cosmology.

    H0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_H0(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``H0``."""
        unit = u.Unit("km/(s Mpc)")

        # on the class
        H0 = cosmo_cls.parameters["H0"]
        assert isinstance(H0, Parameter)
        assert "Hubble constant" in H0.__doc__
        assert H0.unit == unit
        assert H0.default is MISSING

        # validation
        assert H0.validate(cosmo, 1) == 1 * unit
        assert H0.validate(cosmo, 10 * unit) == 10 * unit
        with pytest.raises(ValueError, match="H0 is a non-scalar quantity"):
            H0.validate(cosmo, [1, 2])

        # on the instance
        assert cosmo.H0 is cosmo.__dict__["H0"]
        assert cosmo.H0 == self._cls_args["H0"]
        assert isinstance(cosmo.H0, u.Quantity) and cosmo.H0.unit == unit

    def test_init_H0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``H0``."""
        # test that it works with units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.H0 == ba.arguments["H0"]

        # also without units
        ba.arguments["H0"] = ba.arguments["H0"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.H0.value == ba.arguments["H0"]

        # fails for non-scalar
        ba.arguments["H0"] = u.Quantity([70, 100], u.km / u.s / u.Mpc)
        with pytest.raises(ValueError, match="H0 is a non-scalar quantity"):
            cosmo_cls(*ba.args, **ba.kwargs)


# =============================================================================


class ParameterOm0TestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` Om0 on a Cosmology.

    Om0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Om0(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``Om0``."""
        # on the class
        Om0 = cosmo_cls.parameters["Om0"]
        assert isinstance(Om0, Parameter)
        assert "Omega matter" in Om0.__doc__
        assert Om0.default is MISSING

        # validation
        assert Om0.validate(cosmo, 1) == 1
        assert Om0.validate(cosmo, 10 * u.one) == 10
        with pytest.raises(ValueError, match="Om0 cannot be negative"):
            Om0.validate(cosmo, -1)

        # on the instance
        assert cosmo.Om0 is cosmo.__dict__["Om0"]
        assert cosmo.Om0 == self._cls_args["Om0"]
        assert isinstance(cosmo.Om0, float)

    def test_init_Om0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Om0``."""
        # test that it works with units
        ba.arguments["Om0"] = ba.arguments["Om0"] << u.one  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Om0 == ba.arguments["Om0"]

        # also without units
        ba.arguments["Om0"] = ba.arguments["Om0"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Om0 == ba.arguments["Om0"]

        # fails for negative numbers
        ba.arguments["Om0"] = -0.27
        with pytest.raises(ValueError, match="Om0 cannot be negative."):
            cosmo_cls(*ba.args, **ba.kwargs)


# =============================================================================


class ParameterOde0TestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` Ode0 on a Cosmology.

    Ode0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Parameter_Ode0(self, cosmo_cls: type[Cosmology]):
        """Test Parameter ``Ode0`` on the class."""
        Ode0 = cosmo_cls.parameters.get(
            "Ode0", cosmo_cls._derived_parameters.get("Ode0")
        )
        assert isinstance(Ode0, Parameter)
        assert "Omega dark energy" in Ode0.__doc__
        if issubclass(cosmo_cls, FlatFLRWMixin):
            assert Ode0.default == 0
        else:
            assert Ode0.default is MISSING

    def test_Parameter_Ode0_validation(
        self, cosmo_cls: type[Cosmology], cosmo: Cosmology
    ):
        """Test Parameter ``Ode0`` validation."""
        Ode0 = cosmo_cls.parameters.get(
            "Ode0", cosmo_cls._derived_parameters.get("Ode0")
        )
        assert Ode0.validate(cosmo, 1.1) == 1.1
        assert Ode0.validate(cosmo, 10 * u.one) == 10.0
        with pytest.raises(TypeError, match="only dimensionless"):
            Ode0.validate(cosmo, 10 * u.km)

    def test_Ode0(self, cosmo: Cosmology):
        """Test Parameter ``Ode0`` validation."""
        # if Ode0 is a parameter, test its value
        assert cosmo.Ode0 is cosmo.__dict__["Ode0"]
        assert cosmo.Ode0 == self._cls_args["Ode0"]
        assert isinstance(cosmo.Ode0, float)

    def test_init_Ode0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Ode0``."""
        # test that it works with units
        ba.arguments["Ode0"] = ba.arguments["Ode0"] << u.one  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ode0 == ba.arguments["Ode0"]

        # also without units
        ba.arguments["Ode0"] = ba.arguments["Ode0"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ode0 == ba.arguments["Ode0"]

        # Setting param to 0 respects that. Note this test uses ``Ode()``.
        ba.arguments["Ode0"] = 0.0
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert_quantity_allclose(cosmo.Ode([0, 1, 2, 3]), [0, 0, 0, 0])
        assert_quantity_allclose(cosmo.Ode(1), 0)

        # Must be dimensionless or have no units. Errors otherwise.
        ba.arguments["Ode0"] = 10 * u.km
        with pytest.raises(TypeError, match="only dimensionless"):
            cosmo_cls(*ba.args, **ba.kwargs)


# =============================================================================


class ParameterTcmb0TestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` Tcmb0 on a Cosmology.

    Tcmb0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Tcmb0(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``Tcmb0``."""
        # on the class
        Tcmb0 = cosmo_cls.parameters["Tcmb0"]
        assert isinstance(Tcmb0, Parameter)
        assert "Temperature of the CMB" in Tcmb0.__doc__
        assert Tcmb0.unit == u.K
        assert Tcmb0.default == 0.0 * u.K

        # validation
        assert Tcmb0.validate(cosmo, 1) == 1 * u.K
        assert Tcmb0.validate(cosmo, 10 * u.K) == 10 * u.K
        with pytest.raises(ValueError, match="Tcmb0 is a non-scalar quantity"):
            Tcmb0.validate(cosmo, [1, 2])

        # on the instance
        assert cosmo.Tcmb0 is cosmo.__dict__["Tcmb0"]
        assert cosmo.Tcmb0 == self.cls_kwargs["Tcmb0"]
        assert isinstance(cosmo.Tcmb0, u.Quantity) and cosmo.Tcmb0.unit == u.K

    def test_init_Tcmb0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Tcmb0``."""
        # test that it works with units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Tcmb0 == ba.arguments["Tcmb0"]

        # also without units
        ba.arguments["Tcmb0"] = ba.arguments["Tcmb0"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Tcmb0.value == ba.arguments["Tcmb0"]

        # must be a scalar
        ba.arguments["Tcmb0"] = u.Quantity([0.0, 2], u.K)
        with pytest.raises(ValueError, match="Tcmb0 is a non-scalar quantity"):
            cosmo_cls(*ba.args, **ba.kwargs)


# =============================================================================


class ParameterNeffTestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` Neff on a Cosmology.

    Neff is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Neff(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``Neff``."""
        # on the class
        Neff = cosmo_cls.parameters["Neff"]
        assert isinstance(Neff, Parameter)
        assert "Number of effective neutrino species" in Neff.__doc__
        assert Neff.default == 3.04

        # validation
        assert Neff.validate(cosmo, 1) == 1
        assert Neff.validate(cosmo, 10 * u.one) == 10
        with pytest.raises(ValueError, match="Neff cannot be negative"):
            Neff.validate(cosmo, -1)

        # on the instance
        assert cosmo.Neff is cosmo.__dict__["Neff"]
        assert cosmo.Neff == self.cls_kwargs.get("Neff", 3.04)
        assert isinstance(cosmo.Neff, float)

    def test_init_Neff(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Neff``."""
        # test that it works with units
        ba.arguments["Neff"] = (
            cosmo_cls.parameters["Neff"].default << u.one
        )  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Neff == ba.arguments["Neff"]

        # also without units
        ba.arguments["Neff"] = ba.arguments["Neff"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Neff == ba.arguments["Neff"]

        ba.arguments["Neff"] = -1
        with pytest.raises(ValueError):
            cosmo_cls(*ba.args, **ba.kwargs)


# =============================================================================


class Parameterm_nuTestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` m_nu on a Cosmology.

    m_nu is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_m_nu(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``m_nu``."""
        # on the class
        m_nu = cosmo_cls.parameters["m_nu"]
        assert isinstance(m_nu, Parameter)
        assert "Mass of neutrino species" in m_nu.__doc__
        assert m_nu.unit == u.eV
        assert m_nu.equivalencies == u.mass_energy()
        assert m_nu.default == 0.0 * u.eV

        # on the instance
        # assert cosmo.m_nu is cosmo._m_nu
        assert_quantity_allclose(cosmo.m_nu, [0.0, 0.0, 0.0] * u.eV)

        # set differently depending on the other inputs
        if cosmo.Tnu0.value == 0:
            assert cosmo.m_nu is None
        elif not cosmo._nu_info.has_massive_nu:  # only massless
            assert_quantity_allclose(cosmo.m_nu, 0 * u.eV)
        elif self._nu_info.n_massless_nu == 0:  # only massive
            assert cosmo.m_nu == cosmo._nu_info.nu_y * const.k_B * cosmo.Tnu0
        else:  # a mix -- the most complicated case
            assert_quantity_allclose(
                cosmo.m_nu[: self._nu_info.n_massless_nu], 0 * u.eV
            )
            assert_quantity_allclose(
                cosmo.m_nu[self._nu_info.n_massless_nu],
                cosmo._nu_info.nu_y * const.k_B * cosmo.Tnu0,
            )

    def test_init_m_nu(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``m_nu``.

        Note this requires the class to have a property ``has_massive_nu``.
        """
        # Test that it works when m_nu has units.
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        np.testing.assert_array_equal(
            cosmo.m_nu, cosmo_cls.parameters["m_nu"].default
        )  # (& checks len, unit)
        assert not cosmo.has_massive_nu
        assert cosmo.m_nu.unit == u.eV  # explicitly check unit once.

        # And it works when m_nu doesn't have units.
        ba.arguments["m_nu"] = cosmo_cls.parameters["m_nu"].default.value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        np.testing.assert_array_equal(cosmo.m_nu.value, ba.arguments["m_nu"])
        assert not cosmo.has_massive_nu

        # A negative m_nu raises an exception.
        tba = copy.copy(ba)
        tba.arguments["m_nu"] = u.Quantity([-0.3, 0.2, 0.1], u.eV)
        with pytest.raises(ValueError, match="invalid"):
            cosmo_cls(*tba.args, **tba.kwargs)

    def test_init_m_nu_and_Neff(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``m_nu`` and ``Neff``.

        Note this test requires ``Neff`` as constructor input, and a property
        ``has_massive_nu``.
        """
        # Mismatch with Neff = wrong number of neutrinos
        tba = copy.copy(ba)
        tba.arguments["Neff"] = 4.05
        tba.arguments["m_nu"] = u.Quantity([0.15, 0.2, 0.1], u.eV)
        with pytest.raises(ValueError, match="unexpected number of neutrino"):
            cosmo_cls(*tba.args, **tba.kwargs)

        # No neutrinos, but Neff
        tba.arguments["m_nu"] = 0
        cosmo = cosmo_cls(*tba.args, **tba.kwargs)
        assert not cosmo.has_massive_nu
        assert len(cosmo.m_nu) == 4
        assert cosmo.m_nu.unit == u.eV
        assert_quantity_allclose(cosmo.m_nu, 0 * u.eV)
        # TODO! move this test when create ``test_nu_relative_density``
        assert_quantity_allclose(
            cosmo.nu_relative_density(1.0), 0.22710731766 * 4.05, rtol=1e-6
        )

        # All massive neutrinos case, len from Neff
        tba.arguments["m_nu"] = 0.1 * u.eV
        cosmo = cosmo_cls(*tba.args, **tba.kwargs)
        assert cosmo.has_massive_nu
        assert len(cosmo.m_nu) == 4
        assert cosmo.m_nu.unit == u.eV
        assert_quantity_allclose(cosmo.m_nu, [0.1, 0.1, 0.1, 0.1] * u.eV)

    def test_init_m_nu_override_by_Tcmb0(
        self, cosmo_cls: type[Cosmology], ba: BoundArguments
    ):
        """Test initialization for values of ``m_nu``.

        Note this test requires ``Tcmb0`` as constructor input, and a property
        ``has_massive_nu``.
        """
        # If Neff = 0, m_nu is None.
        tba = copy.copy(ba)
        tba.arguments["Neff"] = 0
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.m_nu is None
        assert not cosmo.has_massive_nu

        # If Tcmb0 = 0, m_nu is None
        tba = copy.copy(ba)
        tba.arguments["Tcmb0"] = 0
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.m_nu is None
        assert not cosmo.has_massive_nu


# =============================================================================


class ParameterOb0TestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` Ob0 on a Cosmology.

    Ob0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Ob0(self, cosmo_cls: type[Cosmology], cosmo: Cosmology):
        """Test Parameter ``Ob0``."""
        # on the class
        Ob0 = cosmo_cls.parameters["Ob0"]
        assert isinstance(Ob0, Parameter)
        assert "Omega baryon;" in Ob0.__doc__
        assert Ob0.default == 0.0

        # validation
        with pytest.warns(DeprecationWarning, match="Ob0=None is deprecated"):
            assert Ob0.validate(cosmo, None) == 0.0
        assert Ob0.validate(cosmo, 0.1) == 0.1
        assert Ob0.validate(cosmo, 0.1 * u.one) == 0.1
        with pytest.raises(ValueError, match="Ob0 cannot be negative"):
            Ob0.validate(cosmo, -1)
        with pytest.raises(ValueError, match="baryonic density can not be larger"):
            Ob0.validate(cosmo, cosmo.Om0 + 1)

        # on the instance
        assert cosmo.Ob0 is cosmo.__dict__["Ob0"]
        assert cosmo.Ob0 == 0.03

    def test_init_Ob0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Ob0``."""
        # test that it works with units
        assert isinstance(ba.arguments["Ob0"], u.Quantity)
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ob0 == ba.arguments["Ob0"]

        # also without units
        ba.arguments["Ob0"] = ba.arguments["Ob0"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ob0 == ba.arguments["Ob0"]

        # Setting param to 0 respects that.  Note this test uses ``Ob()``.
        ba.arguments["Ob0"] = 0.0
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ob0 == 0.0
        if not self.abstract_w:
            assert_quantity_allclose(cosmo.Ob(1), 0)
            assert_quantity_allclose(cosmo.Ob([0, 1, 2, 3]), [0, 0, 0, 0])

        # Negative Ob0 errors
        tba = copy.copy(ba)
        tba.arguments["Ob0"] = -0.04
        with pytest.raises(ValueError, match="Ob0 cannot be negative"):
            cosmo_cls(*tba.args, **tba.kwargs)

        # Ob0 > Om0 errors
        tba.arguments["Ob0"] = tba.arguments["Om0"] + 0.1
        with pytest.raises(ValueError, match="baryonic density can not be larger"):
            cosmo_cls(*tba.args, **tba.kwargs)

        # No baryons specified means baryons = 0 and gives a warning.
        tba = copy.copy(ba)
        tba.arguments["Ob0"] = None
        with pytest.warns(DeprecationWarning, match="Ob0=None is deprecated"):
            cosmo = cosmo_cls(*tba.args, **tba.kwargs)

        # In FLRW `Ob(z)` requires `w(z)`.
        if not self.abstract_w:
            assert cosmo.Ob(1) == 0

        # The default value is None
        assert cosmo_cls.parameters["Ob0"].default == 0.0


# =============================================================================


class ParameterFlatOde0TestMixin(ParameterOde0TestMixin):
    """Tests for `astropy.cosmology.Parameter` Ode0 on a flat Cosmology.

    This will augment or override some tests in ``ParameterOde0TestMixin``.

    Ode0 is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_Parameter_Ode0(self, cosmo_cls: type[Cosmology]):
        """Test Parameter ``Ode0`` on the class."""
        super().test_Parameter_Ode0(cosmo_cls)
        Ode0 = cosmo_cls.parameters.get("Ode0", cosmo_cls._derived_parameters["Ode0"])
        assert Ode0.derived in (True, np.True_)

    def test_Ode0(self, cosmo: Cosmology):
        """Test no-longer-Parameter ``Ode0``."""
        assert cosmo.Ode0 is cosmo.__dict__["Ode0"]
        assert cosmo.Ode0 == 1.0 - (cosmo.Om0 + cosmo.Ogamma0 + cosmo.Onu0)

    def test_init_Ode0(self, cosmo_cls: type[Cosmology], ba: BoundArguments):
        """Test initialization for values of ``Ode0``."""
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.Ode0 == 1.0 - (cosmo.Om0 + cosmo.Ogamma0 + cosmo.Onu0 + cosmo.Ok0)

        # Ode0 is not in the signature
        with pytest.raises(TypeError, match="Ode0"):
            cosmo_cls(*ba.args, **ba.kwargs, Ode0=1)



"""Testing deprecated modules in `astropy.cosmology`."""

import re
import warnings

import pytest


def test_connect():
    """Test `astropy.cosmology.connect`."""
    from astropy.cosmology import connect

    try:
        del connect.CosmologyFromFormat
    except Exception:
        pass

    with (
        warnings.catch_warnings(),  # Always raise warning so 2x test sees it too.
        pytest.deprecated_call(
            match=re.escape("The module `astropy.cosmology.connect` is deprecated"),
        ),
    ):
        warnings.simplefilter("always")
        from astropy.cosmology.connect import CosmologyFromFormat  # noqa: F401


def test_core():
    """Test `astropy.cosmology.core`."""
    from astropy.cosmology import core

    try:
        del core.Cosmology
    except Exception:
        pass

    with (
        warnings.catch_warnings(),  # Always raise warning so 2x test sees it too.
        pytest.deprecated_call(
            match=re.escape("The module `astropy.cosmology.core` is deprecated"),
        ),
    ):
        warnings.simplefilter("always")
        from astropy.cosmology.core import Cosmology  # noqa: F401


def test_flrw():
    """Test `astropy.cosmology.core`."""
    from astropy.cosmology import flrw

    try:
        del flrw.FLRW
    except Exception:
        pass

    with (
        warnings.catch_warnings(),  # Always raise warning so 2x test sees it too.
        pytest.deprecated_call(
            match=re.escape("The module `astropy.cosmology.flrw` is deprecated"),
        ),
    ):
        warnings.simplefilter("always")
        from astropy.cosmology.flrw import FLRW  # noqa: F401


def test_funcs():
    """Test `astropy.cosmology.core`."""
    from astropy.cosmology import funcs

    try:
        del funcs.z_at_value
    except Exception:
        pass

    with (
        warnings.catch_warnings(),  # Always raise warning so 2x test sees it too.
        pytest.deprecated_call(
            match=re.escape("The module `astropy.cosmology.funcs` is deprecated")
        ),
    ):
        warnings.simplefilter("always")
        from astropy.cosmology.funcs import z_at_value  # noqa: F401


def test_parameter():
    """Test `astropy.cosmology.parameter`."""
    from astropy.cosmology import parameter

    try:
        del parameter.Parameter
    except Exception:
        pass

    with (
        warnings.catch_warnings(),  # Always raise warning so 2x test sees it too.
        pytest.deprecated_call(
            match=re.escape("The module `astropy.cosmology.parameter` is deprecated"),
        ),
    ):
        warnings.simplefilter("always")
        from astropy.cosmology.parameter import Parameter  # noqa: F401



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""CMB Temperature.

This is private API. See `~astropy.cosmology.traits` for public API.

"""

__all__ = ["TemperatureCMB"]


from numpy.typing import ArrayLike

from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity


class TemperatureCMB:
    """The trait for computing the cosmological background temperature."""

    Tcmb0: Quantity
    """Temperature of the CMB at z=0."""

    @deprecated_keywords("z", since="7.0")
    def Tcmb(self, z: Quantity | ArrayLike) -> Quantity:
        """Compute the CMB temperature at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Tcmb : Quantity ['temperature']
            The temperature of the CMB.

        Examples
        --------
        >>> import astropy.units as u
        >>> from astropy.cosmology import Planck18, units as cu

        >>> Planck18.Tcmb(u.Quantity([0.5, 1.0], cu.redshift))
        <Quantity [4.08825, 5.451  ] K>

        >>> Planck18.Tcmb(u.Quantity(0.5, ''))
        <Quantity 4.08825 K>

        >>> Planck18.Tcmb(0.5)
        <Quantity 4.08825 K>

        >>> Planck18.Tcmb([0.5, 1.0])
        <Quantity [4.08825, 5.451  ] K>
        """
        return self.Tcmb0 * (aszarr(z) + 1.0)



"""Tests that models are picklable."""

from pickle import dumps, loads

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy import units as u
from astropy.modeling import (
    functional_models,
    mappings,
    math_functions,
    physical_models,
    polynomial,
    powerlaws,
    projections,
    rotations,
    spline,
    tabular,
)
from astropy.modeling.math_functions import ArctanhUfunc
from astropy.utils.compat.optional_deps import HAS_SCIPY

MATH_FUNCTIONS = [func for func in math_functions.__all__ if func != "ArctanhUfunc"]


PROJ_TO_REMOVE = (
    [
        "Projection",
        "Pix2SkyProjection",
        "Sky2PixProjection",
        "Zenithal",
        "Conic",
        "Cylindrical",
        "PseudoCylindrical",
        "PseudoConic",
        "QuadCube",
        "HEALPix",
        "AffineTransformation2D",
        "projcodes",
        "Pix2Sky_ZenithalPerspective",
    ]
    + [f"Pix2Sky_{code}" for code in projections.projcodes]
    + [f"Sky2Pix_{code}" for code in projections.projcodes]
)

PROJECTIONS = [func for func in projections.__all__ if func not in PROJ_TO_REMOVE]

OTHER_MODELS = [
    mappings.Mapping((1, 0)),
    mappings.Identity(2),
    ArctanhUfunc(),
    rotations.Rotation2D(23),
    tabular.Tabular1D(lookup_table=[1, 2, 3, 4]),
    tabular.Tabular2D(lookup_table=[[1, 2, 3, 4], [5, 6, 7, 8]]),
]

POLYNOMIALS_1D = ["Chebyshev1D", "Hermite1D", "Legendre1D", "Polynomial1D"]
POLYNOMIALS_2D = ["Chebyshev2D", "Hermite2D", "Legendre2D", "InverseSIP"]

ROTATIONS = [
    rotations.RotateCelestial2Native(12, 23, 34),
    rotations.RotateNative2Celestial(12, 23, 34),
    rotations.EulerAngleRotation(12, 23, 34, "xyz"),
    rotations.RotationSequence3D([12, 23, 34], axes_order="xyz"),
    rotations.SphericalRotationSequence([12, 23, 34], "xyz"),
    rotations.Rotation2D(12),
]


@pytest.fixture
def inputs():
    return 0.3, 0.4


@pytest.fixture
def inputs_math():
    return 1, -0.5


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
@pytest.mark.parametrize("model", functional_models.__all__)
def test_pickle_functional(inputs, model):
    m = getattr(functional_models, model)()
    mp = loads(dumps(m))
    if m.n_inputs == 1:
        assert_allclose(m(inputs[0]), mp(inputs[0]))
    else:
        assert_allclose(m(*inputs), mp(*inputs))


@pytest.mark.parametrize("model", MATH_FUNCTIONS)
def test_pickle_math_functions(inputs_math, model):
    m = getattr(math_functions, model)()
    mp = loads(dumps(m))
    if m.n_inputs == 1:
        assert_allclose(m(inputs_math[0]), mp(inputs_math[0]))
    else:
        assert_allclose(m(*inputs_math), mp(*inputs_math))


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
@pytest.mark.parametrize("m", OTHER_MODELS)
def test_pickle_other(inputs, m):
    mp = loads(dumps(m))
    if m.n_inputs == 1:
        assert_allclose(m(inputs[0]), mp(inputs[0]))
    else:
        assert_allclose(m(*inputs), mp(*inputs))


def test_pickle_units_mapping(inputs):
    m = mappings.UnitsMapping(((u.m, None),))
    mp = loads(dumps(m))
    assert_allclose(m(inputs[0] * u.km), mp(inputs[0] * u.km))


def test_pickle_affine_transformation_2D(inputs):
    m = projections.AffineTransformation2D(matrix=[[1, 1], [1, 1]], translation=[1, 1])
    m.matrix.fixed = True
    mp = loads(dumps(m))
    assert_allclose(m(*inputs), mp(*inputs))
    assert m.matrix.fixed is True


@pytest.mark.parametrize("model", physical_models.__all__)
def test_pickle_physical_models(inputs, model):
    m = getattr(physical_models, model)()
    m1 = loads(dumps(m))
    if m.n_inputs == 1:
        assert_allclose(m(inputs[0]), m1(inputs[0]))
    else:
        assert_allclose(m(*inputs), m1(*inputs))


@pytest.mark.parametrize("model", POLYNOMIALS_1D)
def test_pickle_1D_polynomials(inputs, model):
    m = getattr(polynomial, model)
    m = m(2)
    m1 = loads(dumps(m))
    assert_allclose(m(inputs[1]), m1(inputs[0]))


@pytest.mark.parametrize("model", POLYNOMIALS_2D)
def test_pickle_2D_polynomials(inputs, model):
    m = getattr(polynomial, model)
    m = m(2, 3)
    m1 = loads(dumps(m))
    assert_allclose(m(*inputs), m1(*inputs))


def test_pickle_polynomial_2D(inputs):
    # Polynomial2D is initialized with 1 degree but
    # requires 2 inputs
    m = polynomial.Polynomial2D
    m = m(2)
    m1 = loads(dumps(m))
    assert_allclose(m(*inputs), m1(*inputs))


def test_pickle_sip(inputs):
    m = polynomial.SIP
    m = m((21, 23), 2, 3)
    m1 = loads(dumps(m))
    assert_allclose(m(*inputs), m1(*inputs))


@pytest.mark.parametrize("model", powerlaws.__all__)
def test_pickle_powerlaws(inputs, model):
    m = getattr(powerlaws, model)()
    m1 = loads(dumps(m))
    if m.n_inputs == 1:
        assert_allclose(m(inputs[0]), m1(inputs[0]))
    else:
        assert_allclose(m(*inputs), m1(*inputs))


@pytest.mark.parametrize("model", PROJECTIONS)
def test_pickle_projections(inputs, model):
    m = getattr(projections, model)()
    m1 = loads(dumps(m))
    assert_allclose(m(*inputs), m1(*inputs))


@pytest.mark.parametrize("m", ROTATIONS)
def test_pickle_rotations(inputs, m):
    mp = loads(dumps(m))

    if m.n_inputs == 2:
        assert_allclose(m(*inputs), mp(*inputs))
    else:
        assert_allclose(m(inputs[0], *inputs), mp(inputs[0], *inputs))


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_pickle_spline(inputs):
    def func(x, noise):
        return np.exp(-(x**2)) + 0.1 * noise

    noise = np.random.randn(50)
    x = np.linspace(-3, 3, 50)
    y = func(x, noise)

    fitter = spline.SplineInterpolateFitter()
    spl = spline.Spline1D(degree=3)
    m = fitter(spl, x, y)

    mp = loads(dumps(m))
    assert_allclose(m(inputs[0]), mp(inputs[0]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Tests for polynomial models."""

# pylint: disable=invalid-name
import os
import unittest.mock as mk
import warnings

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy import conf, wcs
from astropy.io import fits
from astropy.modeling import fitting
from astropy.modeling.functional_models import Linear1D
from astropy.modeling.mappings import Identity
from astropy.modeling.polynomial import (
    SIP,
    Chebyshev1D,
    Chebyshev2D,
    Hermite1D,
    Hermite2D,
    Legendre1D,
    Legendre2D,
    OrthoPolynomialBase,
    Polynomial1D,
    Polynomial2D,
    PolynomialBase,
)
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

linear1d = {
    Chebyshev1D: {
        "args": (3,),
        "kwargs": {"domain": [1, 10]},
        "parameters": {"c0": 1.2, "c1": 2, "c2": 2.3, "c3": 0.2},
        "constraints": {"fixed": {"c0": True}},
    },
    Hermite1D: {
        "args": (3,),
        "kwargs": {"domain": [1, 10]},
        "parameters": {"c0": 1.2, "c1": 2, "c2": 2.3, "c3": 0.2},
        "constraints": {"fixed": {"c0": True}},
    },
    Legendre1D: {
        "args": (3,),
        "kwargs": {"domain": [1, 10]},
        "parameters": {"c0": 1.2, "c1": 2, "c2": 2.3, "c3": 0.2},
        "constraints": {"fixed": {"c0": True}},
    },
    Polynomial1D: {
        "args": (3,),
        "kwargs": {"domain": [1, 10]},
        "parameters": {"c0": 1.2, "c1": 2, "c2": 2.3, "c3": 0.2},
        "constraints": {"fixed": {"c0": True}},
    },
    Linear1D: {
        "args": (),
        "kwargs": {},
        "parameters": {"intercept": 1.2, "slope": 23.1},
        "constraints": {"fixed": {"intercept": True}},
    },
}


linear2d = {
    Chebyshev2D: {
        "args": (1, 1),
        "kwargs": {"x_domain": [0, 99], "y_domain": [0, 82]},
        "parameters": {"c0_0": 1.2, "c1_0": 2, "c0_1": 2.3, "c1_1": 0.2},
        "constraints": {"fixed": {"c0_0": True}},
    },
    Hermite2D: {
        "args": (1, 1),
        "kwargs": {"x_domain": [0, 99], "y_domain": [0, 82]},
        "parameters": {"c0_0": 1.2, "c1_0": 2, "c0_1": 2.3, "c1_1": 0.2},
        "constraints": {"fixed": {"c0_0": True}},
    },
    Legendre2D: {
        "args": (1, 1),
        "kwargs": {"x_domain": [0, 99], "y_domain": [0, 82]},
        "parameters": {"c0_0": 1.2, "c1_0": 2, "c0_1": 2.3, "c1_1": 0.2},
        "constraints": {"fixed": {"c0_0": True}},
    },
    Polynomial2D: {
        "args": (1,),
        "kwargs": {},
        "parameters": {"c0_0": 1.2, "c1_0": 2, "c0_1": 2.3},
        "constraints": {"fixed": {"c0_0": True}},
    },
}

fitters = [
    fitting.LevMarLSQFitter,
    fitting.TRFLSQFitter,
    fitting.LMLSQFitter,
    fitting.DogBoxLSQFitter,
]


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
class TestFitting:
    """Test linear fitter with polynomial models."""

    def setup_class(self):
        self.N = 100
        self.M = 100
        self.x1 = np.linspace(1, 10, 100)
        self.y2, self.x2 = np.mgrid[:100, :83]
        rsn = np.random.default_rng(0)
        self.n1 = rsn.standard_normal(self.x1.size) * 0.1
        self.n2 = rsn.standard_normal(self.x2.size).reshape(self.x2.shape)
        self.linear_fitter = fitting.LinearLSQFitter()

    # TODO: Most of these test cases have some pretty repetitive setup that we
    # could probably factor out

    @pytest.mark.parametrize("model_class", sorted(linear1d, key=str))
    @pytest.mark.parametrize("constraints", [False, True])
    def test_linear_fitter_1D(self, model_class, constraints):
        """Test fitting with LinearLSQFitter"""

        model_args = linear1d[model_class]
        kwargs = {}
        kwargs.update(model_args["kwargs"])
        kwargs.update(model_args["parameters"])

        if constraints:
            kwargs.update(model_args["constraints"])

        model = model_class(*model_args["args"], **kwargs)

        y1 = model(self.x1)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"The fit may be poorly conditioned",
                category=AstropyUserWarning,
            )
            model_lin = self.linear_fitter(model, self.x1, y1 + self.n1)

        if constraints:
            # For the constraints tests we're not checking the overall fit,
            # just that the constraint was maintained
            fixed = model_args["constraints"].get("fixed", None)
            if fixed:
                for param in fixed:
                    expected = model_args["parameters"][param]
                    assert getattr(model_lin, param).value == expected
        else:
            assert_allclose(model_lin.parameters, model.parameters, atol=0.2)

    @pytest.mark.parametrize("model_class", sorted(linear1d, key=str))
    @pytest.mark.parametrize("constraints", [False, True])
    @pytest.mark.parametrize("fitter", fitters)
    def test_non_linear_fitter_1D(self, model_class, constraints, fitter):
        """Test fitting with non-linear LevMarLSQFitter"""
        fitter = fitter()

        model_args = linear1d[model_class]
        kwargs = {}
        kwargs.update(model_args["kwargs"])
        kwargs.update(model_args["parameters"])

        if constraints:
            kwargs.update(model_args["constraints"])

        model = model_class(*model_args["args"], **kwargs)

        y1 = model(self.x1)
        with pytest.warns(AstropyUserWarning, match="Model is linear in parameters"):
            model_nlin = fitter(model, self.x1, y1 + self.n1)

        if constraints:
            fixed = model_args["constraints"].get("fixed", None)
            if fixed:
                for param in fixed:
                    expected = model_args["parameters"][param]
                    assert getattr(model_nlin, param).value == expected
        else:
            assert_allclose(model_nlin.parameters, model.parameters, atol=0.2)

    @pytest.mark.parametrize("model_class", sorted(linear2d, key=str))
    @pytest.mark.parametrize("constraints", [False, True])
    def test_linear_fitter_2D(self, model_class, constraints):
        """Test fitting with LinearLSQFitter"""

        model_args = linear2d[model_class]
        kwargs = {}
        kwargs.update(model_args["kwargs"])
        kwargs.update(model_args["parameters"])

        if constraints:
            kwargs.update(model_args["constraints"])

        model = model_class(*model_args["args"], **kwargs)

        z = model(self.x2, self.y2)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"The fit may be poorly conditioned",
                category=AstropyUserWarning,
            )
            model_lin = self.linear_fitter(model, self.x2, self.y2, z + self.n2)

        if constraints:
            fixed = model_args["constraints"].get("fixed", None)
            if fixed:
                for param in fixed:
                    expected = model_args["parameters"][param]
                    assert getattr(model_lin, param).value == expected
        else:
            assert_allclose(model_lin.parameters, model.parameters, atol=0.2)

    @pytest.mark.parametrize("model_class", sorted(linear2d, key=str))
    @pytest.mark.parametrize("constraints", [False, True])
    @pytest.mark.parametrize("fitter", fitters)
    def test_non_linear_fitter_2D(self, model_class, constraints, fitter):
        """Test fitting with non-linear LevMarLSQFitter"""
        fitter = fitter()

        model_args = linear2d[model_class]
        kwargs = {}
        kwargs.update(model_args["kwargs"])
        kwargs.update(model_args["parameters"])

        if constraints:
            kwargs.update(model_args["constraints"])

        model = model_class(*model_args["args"], **kwargs)

        z = model(self.x2, self.y2)
        with pytest.warns(AstropyUserWarning, match="Model is linear in parameters"):
            model_nlin = fitter(model, self.x2, self.y2, z + self.n2)

        if constraints:
            fixed = model_args["constraints"].get("fixed", None)
            if fixed:
                for param in fixed:
                    expected = model_args["parameters"][param]
                    assert getattr(model_nlin, param).value == expected
        else:
            assert_allclose(model_nlin.parameters, model.parameters, atol=0.2)


@pytest.mark.parametrize("model_class", list(list(linear1d) + list(linear2d)))
def test_polynomial_init_with_constraints(model_class):
    """
    Test that polynomial models can be instantiated with constraints, but no
    parameters specified.

    Regression test for https://github.com/astropy/astropy/issues/3606
    """

    # Just determine which parameter to place a constraint on; it doesn't
    # matter which parameter it is to exhibit the problem so long as it's a
    # valid parameter for the model
    if "1D" in model_class.__name__:
        param = "c0"
    else:
        param = "c0_0"

    if issubclass(model_class, Linear1D):
        param = "intercept"

    if issubclass(model_class, OrthoPolynomialBase):
        degree = (2, 2)
    else:
        degree = (2,)

    m = model_class(*degree, fixed={param: True})

    assert m.fixed[param] is True
    assert getattr(m, param).fixed is True

    if issubclass(model_class, OrthoPolynomialBase):
        assert (
            repr(m)
            == f"<{model_class.__name__}(2, 2, c0_0=0., c1_0=0., c2_0=0., c0_1=0., "
            "c1_1=0., c2_1=0., c0_2=0., c1_2=0., c2_2=0.)>"
        )
        assert (
            str(m) == f"Model: {model_class.__name__}\n"
            "Inputs: ('x', 'y')\n"
            "Outputs: ('z',)\n"
            "Model set size: 1\n"
            "X_Degree: 2\n"
            "Y_Degree: 2\n"
            "Parameters:\n"
            "    c0_0 c1_0 c2_0 c0_1 c1_1 c2_1 c0_2 c1_2 c2_2\n"
            "    ---- ---- ---- ---- ---- ---- ---- ---- ----\n"
            "     0.0  0.0  0.0  0.0  0.0  0.0  0.0  0.0  0.0"
        )
    else:
        if model_class.__name__ == "Polynomial2D":
            assert (
                repr(m) == "<Polynomial2D(2, c0_0=0., c1_0=0., c2_0=0., "
                "c0_1=0., c0_2=0., c1_1=0.)>"
            )
            assert (
                str(m) == "Model: Polynomial2D\n"
                "Inputs: ('x', 'y')\n"
                "Outputs: ('z',)\n"
                "Model set size: 1\n"
                "Degree: 2\n"
                "Parameters:\n"
                "    c0_0 c1_0 c2_0 c0_1 c0_2 c1_1\n"
                "    ---- ---- ---- ---- ---- ----\n"
                "     0.0  0.0  0.0  0.0  0.0  0.0"
            )
        elif model_class.__name__ == "Linear1D":
            assert repr(m) == "<Linear1D(slope=2., intercept=0.)>"
            assert (
                str(m) == "Model: Linear1D\n"
                "Inputs: ('x',)\n"
                "Outputs: ('y',)\n"
                "Model set size: 1\n"
                "Parameters:\n"
                "    slope intercept\n"
                "    ----- ---------\n"
                "      2.0       0.0"
            )
        else:
            assert repr(m) == f"<{model_class.__name__}(2, c0=0., c1=0., c2=0.)>"
            assert (
                str(m) == f"Model: {model_class.__name__}\n"
                "Inputs: ('x',)\n"
                "Outputs: ('y',)\n"
                "Model set size: 1\n"
                "Degree: 2\n"
                "Parameters:\n"
                "     c0  c1  c2\n"
                "    --- --- ---\n"
                "    0.0 0.0 0.0"
            )


def test_sip_hst():
    """Test SIP against astropy.wcs"""

    test_file = get_pkg_data_filename(os.path.join("data", "hst_sip.hdr"))
    hdr = fits.Header.fromtextfile(test_file)
    crpix1 = hdr["CRPIX1"]
    crpix2 = hdr["CRPIX2"]
    wobj = wcs.WCS(hdr)
    a_pars = dict(**hdr["A_*"])
    b_pars = dict(**hdr["B_*"])
    a_order = a_pars.pop("A_ORDER")
    b_order = b_pars.pop("B_ORDER")
    sip = SIP([crpix1, crpix2], a_order, b_order, a_pars, b_pars)
    coords = [1, 1]
    rel_coords = [1 - crpix1, 1 - crpix2]
    astwcs_result = wobj.sip_pix2foc([coords], 1)[0] - rel_coords
    assert_allclose(sip(1, 1), astwcs_result)

    # Test changing of inputs and calling it with keyword argumenrts.
    sip.inputs = ("r", "t")
    assert_allclose(sip(r=1, t=1), astwcs_result)
    assert_allclose(sip(1, t=1), astwcs_result)

    # Test representations
    assert (
        repr(sip) == "<SIP([<Shift(offset=-2048.)>, <Shift(offset=-1024.)>, "
        "<_SIP1D(4, 'A', A_2_0=0.00000855, A_3_0=-0., A_4_0=0., A_0_2=0.00000217, "
        "A_0_3=0., A_0_4=0., A_1_1=-0.0000052, A_1_2=-0., A_1_3=-0., "
        "A_2_1=-0., A_2_2=0., A_3_1=0.)>, "
        "<_SIP1D(4, 'B', B_2_0=-0.00000175, B_3_0=0., B_4_0=-0., B_0_2=-0.00000722, "
        "B_0_3=-0., B_0_4=-0., B_1_1=0.00000618, B_1_2=-0., B_1_3=0., "
        "B_2_1=-0., B_2_2=-0., B_3_1=-0.)>])>"
    )
    with conf.set_temp("max_width", 80):
        # fmt: off
        assert str(sip) == (
            "Model: SIP\n"
            "    Model: Shift\n"
            "    Inputs: ('x',)\n"
            "    Outputs: ('y',)\n"
            "    Model set size: 1\n"
            "    Parameters:\n"
            "         offset\n"
            "        -------\n"
            "        -2048.0\n"
            "\n"
            "    Model: Shift\n"
            "    Inputs: ('x',)\n"
            "    Outputs: ('y',)\n"
            "    Model set size: 1\n"
            "    Parameters:\n"
            "         offset\n"
            "        -------\n"
            "        -1024.0\n"
            "\n"
            "    Model: _SIP1D\n"
            "    Inputs: ('x', 'y')\n"
            "    Outputs: ('z',)\n"
            "    Model set size: 1\n"
            "    Order: 4\n"
            "    Coeff. Prefix: A\n"
            "    Parameters:\n"
            "                A_2_0                 A_3_0          ...         A_3_1        \n"
            "        --------------------- ---------------------- ... ---------------------\n"
            "        8.551277582556502e-06 -4.730444829222791e-10 ... 1.971022971660309e-15\n"
            "\n"
            "    Model: _SIP1D\n"
            "    Inputs: ('x', 'y')\n"
            "    Outputs: ('z',)\n"
            "    Model set size: 1\n"
            "    Order: 4\n"
            "    Coeff. Prefix: B\n"
            "    Parameters:\n"
            "                B_2_0                  B_3_0         ...         B_3_1         \n"
            "        ---------------------- --------------------- ... ----------------------\n"
            "        -1.746491877058669e-06 8.567635427816317e-11 ... -3.779506805487476e-15\n"
        )
        # fmt: on

    # Test get num of coeffs
    assert sip.sip1d_a.get_num_coeff(1) == 6
    # Test error
    MESSAGE = "Degree of polynomial must be 2< deg < 9"
    sip.sip1d_a.order = 1
    with pytest.raises(ValueError, match=MESSAGE):
        sip.sip1d_a.get_num_coeff(1)
    sip.sip1d_a.order = 10
    with pytest.raises(ValueError, match=MESSAGE):
        sip.sip1d_a.get_num_coeff(1)


def test_sip_irac():
    """Test forward and inverse SIP against astropy.wcs"""

    test_file = get_pkg_data_filename(os.path.join("data", "irac_sip.hdr"))
    hdr = fits.Header.fromtextfile(test_file)
    crpix1 = hdr["CRPIX1"]
    crpix2 = hdr["CRPIX2"]
    wobj = wcs.WCS(hdr)
    a_pars = dict(**hdr["A_*"])
    b_pars = dict(**hdr["B_*"])
    ap_pars = dict(**hdr["AP_*"])
    bp_pars = dict(**hdr["BP_*"])
    a_order = a_pars.pop("A_ORDER")
    b_order = b_pars.pop("B_ORDER")
    ap_order = ap_pars.pop("AP_ORDER")
    bp_order = bp_pars.pop("BP_ORDER")
    del a_pars["A_DMAX"]
    del b_pars["B_DMAX"]
    pix = [200, 200]
    rel_pix = [200 - crpix1, 200 - crpix2]
    sip = SIP(
        [crpix1, crpix2],
        a_order,
        b_order,
        a_pars,
        b_pars,
        ap_order=ap_order,
        ap_coeff=ap_pars,
        bp_order=bp_order,
        bp_coeff=bp_pars,
    )

    foc = wobj.sip_pix2foc([pix], 1)
    newpix = wobj.sip_foc2pix(foc, 1)[0]
    assert_allclose(sip(*pix), foc[0] - rel_pix)
    assert_allclose(sip.inverse(*foc[0]) + foc[0] - rel_pix, newpix - pix)

    # Test inverse representations
    assert (
        repr(sip.inverse)
        == "<InverseSIP([<Polynomial2D(2, c0_0=0., c1_0=0.0000114, c2_0=0.00002353, "
        "c0_1=-0.00000546, c0_2=-0.00000667, c1_1=-0.00001801)>, "
        "<Polynomial2D(2, c0_0=0., c1_0=-0.00001495, c2_0=0.00000122, c0_1=0.00001975, "
        "c0_2=-0.00002601, c1_1=0.00002944)>])>"
    )
    assert (
        str(sip.inverse) == "Model: InverseSIP\n"
        "    Model: Polynomial2D\n"
        "    Inputs: ('x', 'y')\n"
        "    Outputs: ('z',)\n"
        "    Model set size: 1\n"
        "    Degree: 2\n"
        "    Parameters:\n"
        "        c0_0   c1_0      c2_0      c0_1       c0_2       c1_1   \n"
        "        ---- -------- --------- ---------- ---------- ----------\n"
        "         0.0 1.14e-05 2.353e-05 -5.463e-06 -6.666e-06 -1.801e-05\n"
        "\n"
        "    Model: Polynomial2D\n"
        "    Inputs: ('x', 'y')\n"
        "    Outputs: ('z',)\n"
        "    Model set size: 1\n"
        "    Degree: 2\n"
        "    Parameters:\n"
        "        c0_0    c1_0       c2_0      c0_1      c0_2       c1_1  \n"
        "        ---- ---------- --------- --------- ---------- ---------\n"
        "         0.0 -1.495e-05 1.225e-06 1.975e-05 -2.601e-05 2.944e-05\n"
    )


def test_sip_no_coeff():
    sip = SIP([10, 12], 2, 2)
    assert_allclose(sip.sip1d_a.parameters, [0.0, 0.0, 0])
    assert_allclose(sip.sip1d_b.parameters, [0.0, 0.0, 0])
    MESSAGE = r"SIP inverse coefficients are not available"
    with pytest.raises(NotImplementedError, match=MESSAGE):
        sip.inverse

    # Test model set
    sip = SIP([10, 12], 2, 2, n_models=2)
    assert sip.sip1d_a.model_set_axis == 0
    assert sip.sip1d_b.model_set_axis == 0


@pytest.mark.parametrize(
    "cls",
    (Polynomial1D, Chebyshev1D, Legendre1D, Polynomial2D, Chebyshev2D, Legendre2D),
)
def test_zero_degree_polynomial(cls):
    """
    A few tests that degree=0 polynomials are correctly evaluated and
    fitted.

    Regression test for https://github.com/astropy/astropy/pull/3589
    """

    MESSAGE = "Degree of polynomial must be positive or null"

    if cls.n_inputs == 1:  # Test 1D polynomials
        p1 = cls(degree=0, c0=1)
        assert p1(0) == 1
        assert np.all(p1(np.zeros(5)) == np.ones(5))

        x = np.linspace(0, 1, 100)
        # Add a little noise along a straight line
        y = 1 + np.random.uniform(0, 0.1, len(x))

        p1_init = cls(degree=0)
        fitter = fitting.LinearLSQFitter()
        p1_fit = fitter(p1_init, x, y)

        # The fit won't be exact of course, but it should get close to within
        # 1%
        assert_allclose(p1_fit.c0, 1, atol=0.10)

        # Error from negative degree
        with pytest.raises(ValueError, match=MESSAGE):
            cls(degree=-1)
    elif cls.n_inputs == 2:  # Test 2D polynomials
        if issubclass(cls, OrthoPolynomialBase):
            p2 = cls(x_degree=0, y_degree=0, c0_0=1)

            # different shaped x and y inputs
            a = np.array([1, 2, 3])
            b = np.array([1, 2])
            with mk.patch.object(
                PolynomialBase,
                "prepare_inputs",
                autospec=True,
                return_value=((a, b), mk.MagicMock()),
            ):
                with pytest.raises(
                    ValueError, match=r"Expected input arrays to have the same shape"
                ):
                    p2.prepare_inputs(mk.MagicMock(), mk.MagicMock())

            # Error from negative degree
            with pytest.raises(ValueError, match=MESSAGE):
                cls(x_degree=-1, y_degree=0)
            with pytest.raises(ValueError, match=MESSAGE):
                cls(x_degree=0, y_degree=-1)
        else:
            p2 = cls(degree=0, c0_0=1)

            # Error from negative degree
            with pytest.raises(ValueError, match=MESSAGE):
                cls(degree=-1)

        assert p2(0, 0) == 1
        assert np.all(p2(np.zeros(5), np.zeros(5)) == np.ones(5))

        y, x = np.mgrid[0:1:100j, 0:1:100j]
        z = (1 + np.random.uniform(0, 0.1, x.size)).reshape(100, 100)

        if issubclass(cls, OrthoPolynomialBase):
            p2_init = cls(x_degree=0, y_degree=0)
        else:
            p2_init = cls(degree=0)
        fitter = fitting.LinearLSQFitter()
        p2_fit = fitter(p2_init, x, y, z)

        assert_allclose(p2_fit.c0_0, 1, atol=0.10)


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
@pytest.mark.parametrize("fitter", fitters)
def test_2d_orthopolynomial_in_compound_model(fitter):
    """
    Ensure that OrthoPolynomialBase (ie. Chebyshev2D & Legendre2D) models get
    evaluated & fitted correctly when part of a compound model.

    Regression test for https://github.com/astropy/astropy/pull/6085.
    """
    fitter = fitter()

    y, x = np.mgrid[0:5, 0:5]
    z = x + y

    simple_model = Chebyshev2D(2, 2)
    with pytest.warns(AstropyUserWarning, match="Model is linear in parameters"):
        simple_fit = fitter(simple_model, x, y, z)

    compound_model = Identity(2) | Chebyshev2D(2, 2)
    compound_model.fittable = True
    compound_model.linear = True
    with pytest.warns(AstropyUserWarning, match="Model is linear in parameters"):
        compound_fit = fitter(compound_model, x, y, z)

    assert_allclose(simple_fit(x, y), compound_fit(x, y), atol=1e-11)


def test_Hermite1D_clenshaw():
    model = Hermite1D(degree=2)

    assert model.clenshaw(1, [3]) == 3
    assert model.clenshaw(1, [3, 4]) == 11
    assert model.clenshaw(1, [3, 4, 5]) == 21
    assert model.clenshaw(1, [3, 4, 5, 6]) == -3


def test__fcache():
    model = OrthoPolynomialBase(x_degree=2, y_degree=2)
    MESSAGE = r"Subclasses should implement this"
    with pytest.raises(NotImplementedError, match=MESSAGE):
        model._fcache(np.asanyarray(1), np.asanyarray(1))

    model = Hermite2D(x_degree=2, y_degree=2)
    assert model._fcache(np.asanyarray(1), np.asanyarray(1)) == {
        0: np.asanyarray(1),
        1: 2,
        3: np.asanyarray(1),
        4: 2,
        2: 2.0,
        5: -4.0,
    }

    model = Legendre2D(x_degree=2, y_degree=2)
    assert model._fcache(np.asanyarray(1), np.asanyarray(1)) == {
        0: np.asanyarray(1),
        1: np.asanyarray(1),
        2: 1.0,
        3: np.asanyarray(1),
        4: np.asanyarray(1),
        5: 1.0,
    }

    model = Chebyshev2D(x_degree=2, y_degree=2)
    assert model._fcache(np.asanyarray(1), np.asanyarray(1)) == {
        0: np.asanyarray(1),
        1: np.asanyarray(1),
        2: 1.0,
        3: np.asanyarray(1),
        4: np.asanyarray(1),
        5: 1.0,
    }


def test_fit_deriv_shape_error():
    model = Hermite2D(x_degree=2, y_degree=2)
    MESSAGE = r"x and y must have the same shape"
    with pytest.raises(ValueError, match=MESSAGE):
        model.fit_deriv(np.array([1, 2]), np.array([3, 4, 5]))

    model = Chebyshev2D(x_degree=2, y_degree=2)
    with pytest.raises(ValueError, match=MESSAGE):
        model.fit_deriv(np.array([1, 2]), np.array([3, 4, 5]))

    model = Legendre2D(x_degree=2, y_degree=2)
    with pytest.raises(ValueError, match=MESSAGE):
        model.fit_deriv(np.array([1, 2]), np.array([3, 4, 5]))

    model = Polynomial2D(degree=2)
    MESSAGE = r"Expected x and y to be of equal size"
    with pytest.raises(ValueError, match=MESSAGE):
        model.fit_deriv(np.array([1, 2]), np.array([3, 4, 5]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astropy cosmology parameter module.

.. deprecated:: 7.1

    This module is deprecated and will be removed in Astropy v8.
    All the public classes and functions have been and will be available
    in the :mod:`~astropy.cosmology` module.

"""

__all__ = ["Parameter"]  # noqa: F822


import sys
import warnings
from typing import Any


def __getattr__(name: str) -> Any:
    if name != "Parameter":
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    from ._src.parameter import Parameter

    setattr(sys.modules[__name__], name, Parameter)

    warnings.warn(
        "The module `astropy.cosmology.parameter` is deprecated since v7.1 and will be "
        "removed in a future version. Import from `astropy.cosmology` instead.",
        category=DeprecationWarning,
    )

    return Parameter



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re

import numpy as np
import pytest

from astropy.time import Time, TimeYearDayTime, conf

iso_times = [
    "2000-02-29",
    "1981-12-31 12:13",
    "1981-12-31 12:13:14",
    "2020-12-31 12:13:14.56",
]
isot_times = [re.sub(" ", "T", tm) for tm in iso_times]
yday_times = ["2000:060", "1981:365:12:13:14", "1981:365:12:13", "2020:366:12:13:14.56"]
# Transpose the array to check that strides are dealt with correctly.
yday_array = np.array(
    [["2000:060", "1981:365:12:13:14"], ["1981:365:12:13", "2020:366:12:13:14.56"]]
).T


def test_fast_conf():
    # Default is to try C parser and then Python parser. Both fail so we get the
    # Python message.
    assert conf.use_fast_parser == "True"  # default
    with pytest.raises(ValueError, match="Time 2000:0601 does not match yday format"):
        Time("2000:0601", format="yday")

    # This is one case where Python parser is different from C parser because the
    # Python parser has a bug and fails with a trailing ".", but C parser works.
    Time("2020:150:12:13:14.", format="yday")
    with conf.set_temp("use_fast_parser", "force"):
        Time("2020:150:12:13:14.", format="yday")
    with conf.set_temp("use_fast_parser", "False"):
        with pytest.raises(ValueError, match="could not convert string to float"):
            Time("2020:150:12:13:14.", format="yday")

    with conf.set_temp("use_fast_parser", "False"):
        assert conf.use_fast_parser == "False"
        # Make sure that this is really giving the Python parser
        with pytest.raises(
            ValueError, match="Time 2000:0601 does not match yday format"
        ):
            Time("2000:0601", format="yday")

    with conf.set_temp("use_fast_parser", "force"):
        assert conf.use_fast_parser == "force"
        # Make sure that this is really giving the Python parser
        err = (
            "fast C time string parser failed: time string ends in middle of component"
        )
        with pytest.raises(ValueError, match=err):
            Time("2000:0601", format="yday")


@pytest.mark.parametrize(
    "times,format",
    [
        (iso_times, "iso"),
        (isot_times, "isot"),
        (yday_times, "yday"),
        (yday_array, "yday"),
    ],
)
@pytest.mark.parametrize("variant", [0, 1, 2])
def test_fast_matches_python(times, format, variant):
    if variant == 0:
        pass  # list/array of different values (null terminated strings included)
    elif variant == 1:
        times = times[-1]  # scalar
    elif variant == 2:
        times = [times[-1]] * 2  # list/array of identical values (no null terminations)

    with conf.set_temp("use_fast_parser", "False"):
        tms_py = Time(times, format=format)

    with conf.set_temp("use_fast_parser", "force"):
        tms_c = Time(times, format=format)

    # Times are binary identical
    assert np.all(tms_py == tms_c)


def test_fast_yday_exceptions():
    # msgs = {1: 'time string ends at beginning of component where break is not allowed',
    #         2: 'time string ends in middle of component',
    #         3: 'required delimiter character not found',
    #         4: 'non-digit found where digit (0-9) required',
    #         5: 'bad day of year (1 <= doy <= 365 or 366 for leap year'}

    with conf.set_temp("use_fast_parser", "force"):
        for times, err in [
            ("2020:150:12", "time string ends at beginning of component"),
            ("2020:150:1", "time string ends in middle of component"),
            ("2020:150*12:13:14", "required delimiter character"),
            ("2020:15*:12:13:14", "non-digit found where digit"),
            ("2020:999:12:13:14", "bad day of year"),
        ]:
            with pytest.raises(ValueError, match=err):
                Time(times, format="yday")


def test_fast_iso_exceptions():
    with conf.set_temp("use_fast_parser", "force"):
        for times, err in [
            ("2020-10-10 12", "time string ends at beginning of component"),
            ("2020-10-10 1", "time string ends in middle of component"),
            ("2020*10-10 12:13:14", "required delimiter character"),
            ("2020-10-10 *2:13:14", "non-digit found where digit"),
        ]:
            with pytest.raises(ValueError, match=err):
                Time(times, format="iso")


def test_fast_non_ascii():
    with pytest.raises(ValueError, match="input is not pure ASCII"):
        with conf.set_temp("use_fast_parser", "force"):
            Time("2020-01-01 1ᛦ:13:14.4324")


def test_fast_subclass():
    """Test subclass where use_fast_parser class attribute is not in __dict__"""

    class TimeYearDayTimeSubClass(TimeYearDayTime):
        name = "yday_subclass"

    # Inheritance works
    assert hasattr(TimeYearDayTimeSubClass, "fast_parser_pars")
    assert "fast_parser_pars" not in TimeYearDayTimeSubClass.__dict__

    try:
        # For YearDayTime, forcing the fast parser with a bad date will give
        # "fast C time string parser failed: time string ends in middle of component".
        # But since YearDayTimeSubClass does not have fast_parser_pars it will
        # use the Python parser.
        with pytest.raises(
            ValueError, match="Time 2000:0601 does not match yday_subclass format"
        ):
            with conf.set_temp("use_fast_parser", "force"):
                Time("2000:0601", format="yday_subclass")
    finally:
        del TimeYearDayTimeSubClass._registry["yday_subclass"]


def test_fast_large_arrays():
    """Test that we do not segfault on large arrays with wrong formats.

    See gh-18254, where this turned out to happen with numpy 2.3.0.
    """
    t = Time(["J2000.0"] * 501)
    assert t.size == 501
    with pytest.raises(ValueError, match="Input values did not match any"):
        Time(["parrot"] * 1000)



import os
import shutil
import sys

import erfa  # noqa: F401
import matplotlib as mpl
import pytest

import astropy  # noqa: F401

if len(sys.argv) == 3 and sys.argv[1] == "--astropy-root":
    ROOT = sys.argv[2]
else:
    # Make sure we don't allow any arguments to be passed - some tests call
    # sys.executable which becomes this script when producing a pyinstaller
    # bundle, but we should just error in this case since this is not the
    # regular Python interpreter.
    if len(sys.argv) > 1:
        print("Extra arguments passed, exiting early")
        sys.exit(1)

for root, dirnames, files in os.walk(os.path.join(ROOT, "astropy")):
    # NOTE: we can't simply use
    # test_root = root.replace('astropy', 'astropy_tests')
    # as we only want to change the one which is for the module, so instead
    # we search for the last occurrence and replace that.
    pos = root.rfind("astropy")
    test_root = root[:pos] + "astropy_tests" + root[pos + 7 :]

    # Copy over the astropy 'tests' directories and their contents
    for dirname in dirnames:
        final_dir = os.path.relpath(os.path.join(test_root, dirname), ROOT)
        # We only copy over 'tests' directories, but not astropy/tests (only
        # astropy/tests/tests) since that is not just a directory with tests.
        if dirname == "tests" and not root.endswith("astropy"):
            shutil.copytree(os.path.join(root, dirname), final_dir, dirs_exist_ok=True)
        else:
            # Create empty __init__.py files so that 'astropy_tests' still
            # behaves like a single package, otherwise pytest gets confused
            # by the different conftest.py files.
            init_filename = os.path.join(final_dir, "__init__.py")
            if not os.path.exists(os.path.join(final_dir, "__init__.py")):
                os.makedirs(final_dir, exist_ok=True)
                with open(os.path.join(final_dir, "__init__.py"), "w") as f:
                    f.write("#")
    # Copy over all conftest.py files
    for file in files:
        if file == "conftest.py":
            final_file = os.path.relpath(os.path.join(test_root, file), ROOT)
            shutil.copy2(os.path.join(root, file), final_file)

# Add the top-level __init__.py file
with open(os.path.join("astropy_tests", "__init__.py"), "w") as f:
    f.write("#")

# Remove test file that tries to import all sub-packages at collection time
os.remove(
    os.path.join("astropy_tests", "utils", "iers", "tests", "test_leap_second.py")
)

# Remove convolution tests for now as there are issues with the loading of the C extension.
# FIXME: one way to fix this would be to migrate the convolution C extension away from using
# ctypes and using the regular extension mechanism instead.
shutil.rmtree(os.path.join("astropy_tests", "convolution"))
os.remove(os.path.join("astropy_tests", "modeling", "tests", "test_convolution.py"))
os.remove(os.path.join("astropy_tests", "modeling", "tests", "test_core.py"))
os.remove(os.path.join("astropy_tests", "visualization", "tests", "test_lupton_rgb.py"))

# FIXME: PIL minversion check does not work
os.remove(
    os.path.join("astropy_tests", "visualization", "wcsaxes", "tests", "test_misc.py")
)
os.remove(
    os.path.join("astropy_tests", "visualization", "wcsaxes", "tests", "test_wcsapi.py")
)

# FIXME: The following tests rely on the fully qualified name of classes which
# don't seem to be the same.
os.remove(os.path.join("astropy_tests", "table", "mixins", "tests", "test_registry.py"))

# Copy the top-level conftest.py
shutil.copy2(
    os.path.join(ROOT, "astropy", "conftest.py"),
    os.path.join("astropy_tests", "conftest.py"),
)

# matplotlib hook in pyinstaller 5.0 and later no longer collects every backend, see
# https://github.com/pyinstaller/pyinstaller/issues/6760
mpl.use("svg")

# We skip a few tests, which are generally ones that rely on explicitly
# checking the name of the current module (which ends up starting with
# astropy_tests rather than astropy).

SKIP_TESTS = [
    "test_exception_logging_origin",
    "test_log",
    "test_configitem",
    "test_config_noastropy_fallback",
    "test_no_home",
    "test_path",
    "test_rename_path",
    "test_data_name_third_party_package",
    "test_pkg_finder",
    "test_wcsapi_extension",
    "test_find_current_module_bundle",
    "test_minversion",
    "test_imports",
    "test_generate_config",
    "test_generate_config2",
    "test_create_config_file",
    "test_download_parallel_fills_cache",
    "test_defined_constants_do_not_change",
    "test_physical_constants_versions",
]

# Run the tests!
sys.exit(
    pytest.main(
        ["astropy_tests", "-k " + " and ".join("not " + test for test in SKIP_TESTS)],
        plugins=[
            "pytest_astropy.plugin",
            "pytest_doctestplus.plugin",
            "pytest_remotedata.plugin",
            "pytest_astropy_header.display",
        ],
    )
)



"""
In this module, we define the coordinate representation classes, which are
used to represent low-level cartesian, spherical, cylindrical, and other
coordinates.
"""

from .base import BaseDifferential, BaseRepresentation, BaseRepresentationOrDifferential
from .cartesian import CartesianDifferential, CartesianRepresentation
from .cylindrical import CylindricalDifferential, CylindricalRepresentation
from .geodetic import (
    BaseBodycentricRepresentation,
    BaseGeodeticRepresentation,
    GRS80GeodeticRepresentation,
    WGS72GeodeticRepresentation,
    WGS84GeodeticRepresentation,
)
from .spherical import (
    BaseSphericalCosLatDifferential,
    BaseSphericalDifferential,
    PhysicsSphericalDifferential,
    PhysicsSphericalRepresentation,
    RadialDifferential,
    RadialRepresentation,
    SphericalCosLatDifferential,
    SphericalDifferential,
    SphericalRepresentation,
    UnitSphericalCosLatDifferential,
    UnitSphericalDifferential,
    UnitSphericalRepresentation,
)

# The following imports are included for backwards compatibility.
# isort: split
from .base import (
    DIFFERENTIAL_CLASSES,
    DUPLICATE_REPRESENTATIONS,
    REPRESENTATION_CLASSES,
    get_reprdiff_cls_hash,
)

__all__ = [
    "BaseBodycentricRepresentation",
    "BaseDifferential",
    "BaseGeodeticRepresentation",
    "BaseRepresentation",
    "BaseRepresentationOrDifferential",
    "BaseSphericalCosLatDifferential",
    "BaseSphericalDifferential",
    "CartesianDifferential",
    "CartesianRepresentation",
    "CylindricalDifferential",
    "CylindricalRepresentation",
    "GRS80GeodeticRepresentation",
    "PhysicsSphericalDifferential",
    "PhysicsSphericalRepresentation",
    "RadialDifferential",
    "RadialRepresentation",
    "SphericalCosLatDifferential",
    "SphericalDifferential",
    "SphericalRepresentation",
    "UnitSphericalCosLatDifferential",
    "UnitSphericalDifferential",
    "UnitSphericalRepresentation",
    "WGS72GeodeticRepresentation",
    "WGS84GeodeticRepresentation",
]



# Script to generate random targets, observatory locations, and times, and
# run these using the Starlink rv command to generate reference values for the
# velocity frame corrections. This requires that Starlink is installed and that
# the rv command is in your PATH. More information about Starlink can be found
# at http://starlink.eao.hawaii.edu/starlink

if __name__ == "__main__":
    from random import choice
    from subprocess import check_output

    import numpy as np

    from astropy import units as u
    from astropy.coordinates import Angle, SkyCoord
    from astropy.table import QTable
    from astropy.time import Time

    np.random.seed(12345)

    N = 100

    tab = QTable()
    target_lon = np.random.uniform(0, 360, N) * u.deg
    target_lat = np.degrees(np.arcsin(np.random.uniform(-1, 1, N))) * u.deg
    tab["target"] = SkyCoord(target_lon, target_lat, frame="fk5")
    tab["obstime"] = Time(
        np.random.uniform(Time("1997-01-01").mjd, Time("2017-12-31").mjd, N),
        format="mjd",
        scale="utc",
    )
    tab["obslon"] = Angle(np.random.uniform(-180, 180, N) * u.deg)
    tab["obslat"] = Angle(np.arcsin(np.random.uniform(-1, 1, N)) * u.deg)
    tab["geocent"] = 0.0
    tab["heliocent"] = 0.0
    tab["lsrk"] = 0.0
    tab["lsrd"] = 0.0
    tab["galactoc"] = 0.0
    tab["localgrp"] = 0.0

    for row in tab:
        # Produce input file for rv command
        with open("rv.input", "w") as f:
            f.write(
                f"{row['obslon'].to_string('deg', sep=' ')}"
                f" {row['obslat'].to_string('deg', sep=' ')}\n"
            )
            f.write(
                f"{row['obstime'].datetime.year} {row['obstime'].datetime.month}"
                f" {row['obstime'].datetime.day} 1\n"
            )
            f.write(row["target"].to_string("hmsdms", sep=" ") + " J2000\n")
            f.write("END\n")

        # Run Starlink rv command
        check_output(["rv", "rv.input"])

        # Parse values from output file
        lis_lines = []
        started = False
        for lis_line in open("rv.lis"):
            if started and lis_line.strip() != "":
                lis_lines.append(lis_line.strip())
            elif "LOCAL GROUP" in lis_line:
                started = True

        # Some sources are not observable at the specified time and therefore don't
        # have entries in the rv output file
        if len(lis_lines) == 0:
            continue

        # If there are lines, we pick one at random. Note that we can't get rv to
        # run at the exact time we specified in the input, so we will re-parse the
        # actual date/time used and replace it in the table
        lis_line = choice(lis_lines)

        # The column for 'SUN' has an entry also for the light travel time, which
        # we want to ignore. It sometimes includes '(' followed by a space which
        # can cause issues with splitting, hence why we get rid of the space.
        lis_line = lis_line.replace("(  ", "(").replace("( ", "(")
        (
            year,
            month,
            day,
            time,
            zd,
            row["geocent"],
            row["heliocent"],
            _,
            row["lsrk"],
            row["lsrd"],
            row["galactoc"],
            row["localgrp"],
        ) = lis_line.split()
        row["obstime"] = Time(
            f"{year}-{month}-{day}T{time}:00", format="isot", scale="utc"
        )

    # We sampled 100 coordinates above since some may not have results - we now
    # truncate to 50 sources since this is sufficient.
    tab[:50].write("reference_rv.ecsv", format="ascii.ecsv")
